# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""The main module: sets global variables and registers OmegaConf resolvers."""

import uuid
import git
from omegaconf import OmegaConf
import skais_mapper.__about__
import skais_mapper.rotations
import skais_mapper.cosmology
import skais_mapper.raytrace
import skais_mapper.illustris
import skais_mapper.simobjects
import skais_mapper.plotting
import skais_mapper.utils
import skais_mapper.configure
import skais_mapper.generate


RUN_UID = uuid.uuid4()
repository = git.Repo(search_parent_directories=True)
GIT_STATE = repository.head.object.hexsha
GIT_DIFF = [
    str(diff)
    for diff in
    repository.index.diff(None, create_patch=True)
    + repository.index.diff("HEAD", create_patch=True)
]

if OmegaConf.has_resolver("get_run_id"):
    OmegaConf.clear_resolver("get_run_id")
OmegaConf.register_new_resolver("get_run_id", skais_mapper.utils.get_run_id)



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Plotting routines for maps and images."""

from pathlib import Path
import numpy as np
import astropy.units as au
from skais_mapper.utils import SkaisColorMaps, get_run_id, alias_kw
from matplotlib import colormaps
from matplotlib import pyplot as plt
from matplotlib.image import AxesImage
from matplotlib.colors import Colormap
from functools import singledispatch
from typing import Any
from collections.abc import Sequence

try:
    import torch
    from torch import Tensor

    TORCH_INSTALLED = True
except ImportError:
    TORCH_INSTALLED = False

    class Tensor:
        """Dummy Tensor class for typing."""

        ...


def _from_batch(
    batch: torch.Tensor | np.ndarray | Sequence[torch.Tensor | np.ndarray],
    metadata: dict | Sequence[dict] = {},
    batch_idx: int | Sequence[int] | None = None,
) -> tuple[torch.Tensor, dict | None]:
    """Select data from batch using appropriate batch indexing.

    Args:
        batch: Data batch to be indexed.
        metadata: Corresponding metadata dictionaries.
        batch_idx: If not `None`, data is a single or multiple batches of images
    """
    data = batch
    if batch_idx is not None:
        batch_idx = (batch_idx,) if isinstance(batch_idx, int) else batch_idx
        for idx in batch_idx:
            data = data[idx]
            if isinstance(metadata, Sequence):
                metadata = metadata[idx]
        if isinstance(metadata, dict):
            for key in metadata:
                if isinstance(metadata[key], Tensor):
                    metadata[key] = metadata[key].detach().clone()[batch_idx[-1]]
                if isinstance(metadata[key], np.ndarray | list | tuple):
                    metadata[key] = metadata[key][batch_idx[-1]]
    if len(data.shape) >= 2:
        data = data.squeeze()
    return data, metadata


def _get_cmap(
    name: str | None = None,
    type_: str | None = None,
    under: str | None = None,
    over: str | None = None,
    bad: str | None = None,
) -> Colormap:
    """Get colormap by data map type or by name directly.

    Args:
        name: Name of the colormap (names from SkaisColorMaps or matplotlib allowed).
        type_: Type of the map.
        under: If not `None`, sets the colors below the colormap.
        over: If not `None`, sets the colors above the colormap.
        bad: If not `None`, sets the colormap colors for bad values.
    """
    if type_ is not None:
        match type_:
            case "gas":
                cmap = "gaseous"
            case "dm":
                cmap = "obscura"
            case "star":
                cmap = "hertzsprung"
            case "bfield":
                cmap = "gravic"
            case "temp":
                cmap = "phoenix"
            case "hi/21cm":
                cmap = "nava"
            case _:
                cmap = "gaseous"
    elif name is not None:
        if hasattr(SkaisColorMaps, name):
            cmap = name
        elif name in list(colormaps):
            cmap = name
        else:
            cmap = "gaseous"
    else:
        cmap = "gaseous"
    cmap = getattr(SkaisColorMaps, cmap) if hasattr(SkaisColorMaps, cmap) else plt.get_cmap(cmap)
    if under is not None:
        cmap.set_under(under)
    if over is not None:
        cmap.set_under(over)
    if bad is not None:
        cmap.set_bad(bad)
    return cmap


def _symbol_from_class(class_type: str | None = None):
    """Get physical/mathematical symbol (LaTeX notation) from class string.

    Args:
        class_type: Quantity name; one of
          ["dm", "star", "gas", "hi", "hi/21cm", "temp", "bfield"]
    """
    match class_type:
        case "gas":
            symbol = "\u03a3" + r"$_{\mathrm{gas}}$"
        case "hi":
            symbol = "\u03a3" + r"$_{\mathrm{HI}}$"
        case "dm":
            symbol = "\u03a3" + r"$_{\mathrm{dm}}$"
        case "star":
            symbol = "\u03a3" + r"$_{\mathrm{star}}$"
        case "bfield":
            symbol = "|B|"
        case "temp":
            symbol = "T"
        case "hi/21cm":
            symbol = r"T$_{\mathrm{b}}$"
        case _:
            symbol = "\u03a3"
    return symbol


@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("colorbar_label", "cbar_label")
def _plot_data(
    data: np.ndarray,
    info: dict = {},
    extent: Sequence[float] | None = None,
    colormap: None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
) -> AxesImage:
    """Plot image data array.

    Args:
        data: Array(s) of a map (or maps).
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    figsize = kwargs.pop("figsize", [6.4, 4.8])
    dpi = kwargs.pop("dpi", 200)
    transparent = kwargs.pop("transparent", True)
    bbox_inches = kwargs.pop("bbox_inches", "tight")
    pad_inches = kwargs.pop("pad_inches", 0)
    kwargs.setdefault("interpolation", "bicubic")
    kwargs.setdefault("origin", "lower")
    colormap = _get_cmap(colormap, info["class"] if "class" in info else None, under="k", bad="k")
    if "extent" in info and extent is None:
        extent = info["extent"]
    plt.figure(figsize=figsize, dpi=dpi)
    img = plt.imshow(data, cmap=colormap, extent=extent, **kwargs)
    if colorbar:
        if colorbar_label is None:
            colorbar_label = _symbol_from_class(info.get("class", None))
            if "units" in info:
                colorbar_label += f" [{info['units']}]"
        colorbar_label = colorbar_label.replace("solMass", "M" + r"$_{\odot}$").replace(
            "2", "\u00b2"
        )
        plt.colorbar(label=colorbar_label)
    if "units_extent" in info and xlabel is None:
        xlabel = f"[{info['units_extent']}]"
    if "units_extent" in info and ylabel is None:
        ylabel = f"[{info['units_extent']}]"
    if xlabel is not None:
        plt.xlabel(xlabel)
    if ylabel is not None:
        plt.ylabel(ylabel)
    if savefig:
        if path is None:
            path = f"./{get_run_id()}_image.png"
        path = Path(path)
        if not path.parent.exists():
            path.parent.mkdir(parents=True)
        plt.savefig(
            path, dpi=dpi, bbox_inches=bbox_inches, pad_inches=pad_inches, transparent=transparent
        )
        if verbose:
            print(f"Saving to [png]: {path}")
        if close:
            plt.close()
    if show:
        plt.show()
    return img


@singledispatch
@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("colorbar_label", "cbar_label")
def plot_image(
    data: Any,
    batch_idx: int | None = None,
    extent: Sequence[float] | None = None,
    colormap: None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
):
    """Plot map from an array, tensor, or batch(es).

    Args:
        data: Array(s) of a map (or maps).
        batch_idx: If not `None`, data is a single or multiple batches of images.
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    raise NotImplementedError(f"Invalid data type {type(data)}.")


@plot_image.register(Tensor)
@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("colorbar_label", "cbar_label")
def plot_image_tensor(
    data: Tensor,
    batch_idx: int | Sequence[int] | None = None,
    extent: Sequence[float] | None = None,
    colormap: Colormap | None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
):
    """Plot map from a tensor, or batch(es).

    Args:
        data: Array(s) of a map (or maps)
        batch_idx: If not `None`, data is a single or multiple batches of images.
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    metadata = kwargs.pop("info", {})
    data, metadata = _from_batch(data, metadata=metadata, batch_idx=batch_idx)
    _plot_data(
        data,
        metadata,
        extent=extent,
        colormap=colormap,
        colorbar=colorbar,
        colorbar_label=colorbar_label,
        xlabel=xlabel,
        ylabel=ylabel,
        savefig=savefig,
        path=path,
        show=show,
        close=close,
        verbose=verbose,
        **kwargs,
    )


@plot_image.register(np.ndarray)
@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("cbar_label", "colorbar_label")
def plot_image_array(
    data: np.ndarray,
    batch_idx: int | Sequence[int] | None = None,
    extent: Sequence[float] | None = None,
    colormap: Colormap | None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
):
    """Plot map from an array or batch(es).

    Args:
        data: Array(s) of a map (or maps)
        batch_idx: If not `None`, data is a single or multiple batches of images.
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    metadata = kwargs.pop("info", {})
    data, metadata = _from_batch(data, metadata=metadata, batch_idx=batch_idx)
    _plot_data(
        data,
        metadata,
        extent=extent,
        colormap=colormap,
        colorbar=colorbar,
        colorbar_label=colorbar_label,
        xlabel=xlabel,
        ylabel=ylabel,
        savefig=savefig,
        path=path,
        show=show,
        close=close,
        verbose=verbose,
        **kwargs,
    )


@plot_image.register(au.Quantity)
@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("cbar_label", "colorbar_label")
def plot_image_quantity(
    data: au.Quantity,
    batch_idx: int | Sequence[int] | None = None,
    extent: Sequence[float] | None = None,
    colormap: Colormap | None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
):
    """Plot map from an astropy.units.Quantity array or batch(es).

    Args:
        data: Array(s) of a map (or maps)
        batch_idx: If not `None`, data is a single or multiple batches of images.
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    metadata = kwargs.pop("info", {})
    data, metadata = _from_batch(data, metadata=metadata, batch_idx=batch_idx)
    data, units = data.value, data.unit
    if "units" not in metadata:
        metadata["units"] = f"{units:latex}"
    _plot_data(
        data,
        metadata,
        extent=extent,
        colormap=colormap,
        colorbar=colorbar,
        colorbar_label=colorbar_label,
        xlabel=xlabel,
        ylabel=ylabel,
        savefig=savefig,
        path=path,
        show=show,
        close=close,
        verbose=verbose,
        **kwargs,
    )


@plot_image.register(Sequence)
@alias_kw("colormap", "cmap")
@alias_kw("colorbar", "cbar")
@alias_kw("cbar_label", "colorbar_label")
def plot_image_from_batch(
    data: Sequence[Tensor | np.ndarray | dict],
    batch_idx: int | Sequence[int] | None = None,
    extent: Sequence[float] | None = None,
    colormap: None = None,
    colorbar: bool = False,
    colorbar_label: str | None = None,
    xlabel: str | None = None,
    ylabel: str | None = None,
    savefig: bool = False,
    path: str | Path | None = None,
    show: bool = False,
    close: bool = False,
    verbose: bool = False,
    **kwargs,
):
    """Plot map from an array/tensor batch (optionally including metadata).

    Args:
        data: Array(s) of a map (or maps)
        batch_idx: If not `None`, data is a single or multiple batches of images.
        info: Supplementary information about the data.
        extent: Extent of the map in physical units.
        colormap: Name of the colormap to use for plotting.
        colorbar: If `True`, include the colorbar in the plot.
        colorbar_label: Label of the colorbar (if enabled).
        xlabel: Label of the x-axis.
        ylabel: Label of the y-axis.
        savefig: If `True`, save the plot to file.
        path: Filename or filepath where the figure is saved.
        show: If `True`, show the plot.
        close: If `True`, closes the figure.
        verbose: Print results to stdout.
        kwargs: Additional keyword arguments for `matplotlib.pyplot.imshow`.
    """
    if all([isinstance(d, Tensor) or isinstance(d, np.ndarray) for d in data]):
        data = data
        metadata = {}
    else:
        data, metadata = data
    metadata = kwargs.pop("info", metadata)
    data, metadata = _from_batch(data, metadata, batch_idx=batch_idx)
    _plot_data(
        data,
        metadata,
        extent=extent,
        colormap=colormap,
        colorbar=colorbar,
        colorbar_label=colorbar_label,
        xlabel=xlabel,
        ylabel=ylabel,
        savefig=savefig,
        path=path,
        show=show,
        close=close,
        verbose=verbose,
        **kwargs,
    )


if __name__ == "__main__":
    data_img = torch.rand((1, 512, 512))
    data_imgs = torch.rand((5, 1, 512, 512))
    data_arr = np.random.rand(5, 1, 512, 512)
    data_qty = au.Quantity(data_arr, "m/s")
    data_batch = (torch.rand((5, 1, 512, 512)), torch.rand((5, 1, 512, 512)))
    batch = (
        (torch.rand((5, 1, 512, 512)), torch.rand((5, 1, 512, 512))),
        (
            {
                "class": ["gas"],
                "extent": [[-42.0, 42.0, -42.0, 42.0]],
                "units_extent": ["kpc"],
            },
            {
                "class": ["dm"],
                "extent": [[-91.0, 91.0, -91.0, 91.0]],
                "units_extent": ["kpc"],
            },
        ),
    )
    plot_image(data_img)
    plot_image(data_imgs, 0)
    plot_image(data_arr, 2)
    plot_image(data_qty, 2)
    plot_image(data_batch, (1, 3))
    plot_image(batch, (1, 0))



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Color module for more beautiful plots."""

import random
import numpy as np
from matplotlib.colors import Colormap, LinearSegmentedColormap, Normalize, to_hex
from matplotlib.cm import ScalarMappable
from matplotlib import pyplot as plt

__all__ = ["color_variant", "SkaisColors", "SkaisColorMaps"]


def color_variant(hex_color: str, shift: int = 10) -> str:
    """Takes a color in hex code and produces a lighter or darker shift variant.

    Args:
        hex_color (str): formatted as '#' + rgb hex string of length 6
        shift (int): decimal shift of the rgb hex string

    Returns:
        variant (str): formatted as '#' + rgb hex string of length 6
    """
    if len(hex_color) != 7:
        message = "Passed {} to color_variant(), needs to be in hex format."
        raise ValueError(message.format(hex_color))
    rgb_hex = [hex_color[x : x + 2] for x in [1, 3, 5]]
    new_rgb_int = [int(hex_value, 16) + shift for hex_value in rgb_hex]
    # limit to interval 0 and 255
    new_rgb_int = [min([255, max([0, i])]) for i in new_rgb_int]
    # hex() produces "0x88", we want the last two digits
    return "#" + "".join([hex(i)[2:] for i in new_rgb_int])


class ReNormColormapAdaptor(Colormap):
    """Adaptor for re-normalizing color mappable."""

    def __init__(self, base, cmap_norm, orig_norm=None):
        if orig_norm is None:
            if isinstance(base, ScalarMappable):
                orig_norm = base.norm
                base = base.cmap
            else:
                orig_norm = Normalize(0, 1)
        self._base = base
        if isinstance(cmap_norm, type(Normalize)) and issubclass(cmap_norm, Normalize):
            # a class was provided instead of an instance. create an instance
            # with the same limits.
            cmap_norm = cmap_norm(orig_norm.vmin, orig_norm.vmax)
        self._cmap_norm = cmap_norm
        self._orig_norm = orig_norm

    def __call__(self, X, **kwargs):
        """Re-normalise the values before applying the colormap."""
        return self._base(self._cmap_norm(self._orig_norm.inverse(X)), **kwargs)

    def __getattr__(self, attr):
        """Any other attribute, we simply dispatch to the underlying cmap."""
        return getattr(self._base, attr)

    def _init(self, *args, **kwargs):
        """Abstract dummy method."""

    def reversed(self, *args, **kwargs):
        """Abstract dummy method."""


class SkaisColors:
    """An assortment of colors and palettes."""

    # Shades
    white = "#DDDEE1"  # rba(221, 222, 225)
    gray = "#98989D"  # rba(152, 152, 157)
    grey = "#98989D"  # rba(152, 152, 157)
    darkish = "#666769"  # rba(102, 103, 105)
    dark = "#3d3e41"  # rba( 61,  62,  65)
    darker = "#333437"  # rba( 51,  52,  55)
    darkest = "#212225"  # rba( 33,  34,  37)
    black = "#090F0F"  # rba(  9,  15,  15)
    textcolor = "#dddee1"  # rba(221, 222, 225)
    # Primary colors
    red = "#FF6767"
    pink = "#FF375F"  # rba(255,  55,  95)
    orange = "#FF9F0A"  # rba(255, 159,  10)
    yellow = "#FFD60A"  # rba(155, 214,  10)
    purple = "#603DD0"  # rba( 96,  61, 208)
    green = "#32D74B"  # rba( 50, 215,  75)
    cyan = "#5BC1AE"
    blue = "#6767FF"
    brown = "#D88C4E"  # rba(172, 142, 104)
    # Other
    golden = "#feb125"  # rba(256, 177,  37)
    purpleblue = "#7d7de1"  # rba(125, 125, 225)
    turquoise = "#00d1a4"  # rba( 10, 210, 165)
    # Light colors
    cyan_light = "#A0DED2"
    light_cloors = [cyan_light]
    # Dark colors
    cyan_dark = "#24A38B"
    blue_marguerite = "#756BB1"
    dark_colors = [cyan_dark]
    misc_colors = [blue_marguerite]

    colors = [
        red,
        pink,
        orange,
        yellow,
        purple,
        green,
        cyan,
        blue,
        brown,
        white,
        grey,
        cyan_light,
        cyan_dark,
    ]

    # plotting palettes
    plt5 = ["#ffa600", "#ff6361", "#bc5090", "#58508d", "#668B9d"]
    plt8 = ["#006999", "#4d6db1", "#8b6bbb", "#c464b2", "#f15e99", "#ff6773", "#ff8246", "#ffa600"]
    # Scheme colors
    palettes = {
        "arcus": ["#FC354C", "#29221F", "#13747D", "#0ABFBC", "#FCF7C5"],
        "aquaria": ["#00207F", "#A992FA", "#EA55B1", "#FEC763"],
        "geometric": ["#08F7FE", "#09FBD3", "#FE53BB", "#F5D300"],
        "neon": ["#560A86", "#7122FA", "#F148FB", "#FFACFC"],
        "psychedelic": ["#011FFD", "#FF2281", "#B76CFD", "#75D5FD"],
        "vivid": ["#FDC7D7", "#FF9DE6", "#A5D8F3", "#E8E500"],
        "abstract": ["#7B61F8", "#FF85EA", "#FDF200", "#00FECA"],
        "phoenix": ["#33135C", "#652EC7", "#DE38C8", "#FFC300"],
        "cosmicnova": ["#3B27BA", "#E847AE", "#13CA91", "#FF9472"],
        "pinkpop": ["#35212A", "#3B55CE", "#FF61BE", "#FFDEF3"],
        "agaveglitch": ["#01535F", "#02B8A2", "#FDB232", "#FDD400"],
        "coralglow": ["#F85125", "#FF8B8B", "#FEA0FE", "#79FFFE"],
        "luxuria": ["#037A90", "#00C2BA", "#FF8FCF", "#CE96FB"],
        "luminos": ["#9D72FF", "#FFB3FD", "#01FFFF", "#01FFC3"],
        "stationary": ["#FE6B25", "#28CF75", "#EBF875", "#A0EDFF"],
        "prism": ["#C24CF6", "#FF1493", "#FC6E22", "#FFFF66"],
        "retro": ["#CE0000", "#FF5F01", "#FE1C80", "#FFE3F1"],
        "terrean": ["#E8DDCB", "#CDB380", "#036564", "#033649", "#031634"],
        "cozytime": ["#FFEEB2", "#F4CE61", "#FFBF20", "#7D9CA1", "#4D8D97"],
        "gravic": ["#0b0b13", "#272e59", "#4d5693", "#ffbf20", "#f8f5ec"],
        "marsian": ["#e8ddcb", "#cd967e", "#63031d", "#490303", "#1d0202"],
        "twofourone": ["#D1313D", "#E5625C", "#F9BF76", "#8EB2C5", "#615375"],
        "ioba": ["#FF3366", "#C74066", "#8F4D65", "#575A65", "#1F6764"],
        "mintrock": ["#595B5A", "#14C3A2", "#0DE5A8"],
        "pukedrainbow": ["#482344", "#2B5166", "#429867", "#FAB243", "#E02130"],
        "acryliq": ["#F21A1D", "#FF822E", "#03DDDC", "#FEF900"],
        "hibokeh": ["#0310EA", "#FB33DB", "#7FFF00", "#FCF340"],
        "flashy": ["#04005E", "#440BD4", "#FF2079", "#E92EFB"],
        "cyber": ["#00218A", "#535EEB", "#BC75F9", "#BDBDFD"],
        "cyberfade": ["#00218A", "#535EEB", "#BC75F9", "#BDBDFD", "#DEDEFE", "#FFFFFF"],
        "zoas": ["#001437", "#7898FB", "#5CE5D5", "#B8FB3C"],
        "vilux": ["#001437", "#85B2FF", "#17E7B6", "#D4FD87", "#FDEB87", "#FDD74C", "#FCAA43"],
        "graphiq": ["#48ADF1", "#C6BDEA", "#FDCBFC", "#8AF7E4"],
        "vibes": ["#027A9F", "#12B296", "#FFAA01", "#E1EF7E"],
        "purplerain": ["#120052", "#8A2BE2", "#B537F2", "#3CB9FC"],
        "gaseous": [
            "#101016",
            "#1B366A",
            "#2F4F85",
            "#385B8F",
            "#79695d",
            "#D98647",
            "#EBC79A",
            "#FFFFFF",
        ],
        "nava": [
            "#0A001B",
            "#3A2B62",
            "#72678F",
            "#5C93A7",
            "#749DBB",
            "#84D7E7",
            "#ABF2EC",
            "#EEF5FF",
            "#FFFFFF",
            "#FCF9FF",
            "#F19B8A",
            "#D93E44",
            "#C7204C",
            "#18001a",
        ],
        "obscura": [
            "#101016",
            "#103F39",
            "#27524C",
            "#52746F",
            "#6F795D",
            "#D98647",
            "#EBC79A",
            "#FFFFFF",
        ],
        "hertzsprung": [
            "#101016",
            "#6A1B36",
            "#852F4F",
            "#8F385B",
            "#D98647",
            "#EBC79A",
            "#FFFFFF",
        ],
        "shapely": [
            (30.0 / 255, 136.0 / 255, 229.0 / 255, alpha) for alpha in np.linspace(1, 0, 100)
        ]
        + [(255.0 / 255, 13.0 / 255, 87.0 / 255, alpha) for alpha in np.linspace(0, 1, 100)],
    }

    for k, p in palettes.items():
        locals()[k + "_palette"] = p
        for i, pi in enumerate(p):
            locals()[k + f"{i}"] = pi
    locals().pop("k")
    locals().pop("p")
    locals().pop("i")
    locals().pop("pi")

    @classmethod
    def cmap_from_color(
        cls,
        color_str: str,
        secondary_color: str | None = None,
    ) -> LinearSegmentedColormap:
        """Create a colormap from a single color.

        Args:
            color_str: color string of the class color
            secondary_color: color into which the color changes in the colormap

        Returns:
            (mpl.colors.LinearSegmentedColormap object): reversed colormap
        """
        if secondary_color is None:
            secondary_color = color_variant(color_str, shift=125)
        if color_str in cls.__dict__:
            color = cls.__dict__[color_str]
        else:
            color = color_str
        cmap = LinearSegmentedColormap.from_list("Skais" + color_str, [secondary_color, color])
        return cmap


class SkaisColorMaps:
    """An assortment of linearly interpolated colormaps based on 4+ colors each."""

    for k, p in SkaisColors.palettes.items():
        locals()[k] = LinearSegmentedColormap.from_list(k, p)
    aslist = []
    for k in SkaisColors.palettes:
        aslist.append(locals()[k])
    asarray = np.asarray(aslist)
    N = len(aslist)
    locals().pop("k")
    locals().pop("p")

    @classmethod
    def random(cls) -> LinearSegmentedColormap:
        """Choose a random color map.

        Returns:
            cmap: random colormap from custom list
        """
        return random.choice(cls.aslist)

    @classmethod
    def gen(cls):
        """Generate colormaps.

        Returns:
            (mpl.colors.LinearSegmentedColormap object): colormap generated from custom list
        """
        yield from cls.aslist

    @classmethod
    def reverse(
        cls, cmap: Colormap, set_bad: str = None, set_under: str = None, set_over: str = None
    ) -> LinearSegmentedColormap:
        """Reverse the specified colormap.

        Args:
            cmap (mpl.colors.LinearSegmentedColormap object): colormap to be reversed
            set_bad (str): set colormaps bad values to a different color
            set_under (str): set colormaps under values to a different color
            set_over (str): set colormaps over values to a different color

        Returns:
            (mpl.colors.LinearSegmentedColormap object): reversed colormap
        """
        reverse = []
        k = []

        for key in cmap._segmentdata:
            k.append(key)

            channel = cmap._segmentdata[key]
            data = []

            for t in channel:
                data.append((1 - t[0], t[2], t[1]))
            reverse.append(sorted(data))

        linear_l = dict(zip(k, reverse))
        rcmap = LinearSegmentedColormap(f"{cmap.name}_r", linear_l)
        if set_bad is not None:
            rcmap.set_bad(set_bad)
        if set_under is not None:
            rcmap.set_over(set_under)
        if set_over is not None:
            rcmap.set_over(set_over)
        return rcmap

    @classmethod
    def palette(cls, cmap_name: str, N: int):
        """Return a palette of a colormap with N linearly interpolated color points.

        Args:
            cmap_name: name of the colormap
            N: number of colors in the palette
        """
        vals = np.linspace(0, 1, N)
        cmap = cls.__dict__[cmap_name]
        rgbas = cmap(vals)
        return [to_hex(rgba, keep_alpha=False) for rgba in rgbas]

    @classmethod
    def plot_gradients(cls, savefig: bool = False):
        """Plot all color-map gradients.

        Args:
            savefig (bool): save figure as palettes.png
        """
        gradient = np.linspace(0, 1, 256)
        gradient = np.vstack((gradient, gradient))
        fig, axes = plt.subplots(nrows=SkaisColorMaps.N)
        fig.subplots_adjust(top=0.95, bottom=0.01, left=0.2, right=0.99)
        for ax, cmap in zip(axes, cls.aslist):
            ax.imshow(gradient, aspect="auto", cmap=cmap)
            pos = list(ax.get_position().bounds)
            x_text = pos[0] - 0.01
            y_text = pos[1] + pos[3] / 2.0
            fig.text(x_text, y_text, cmap.name, va="center", ha="right", fontsize=10)
        for ax in axes:
            ax.set_axis_off()
        if savefig:
            plt.savefig("palette.png")

    @staticmethod
    def register_all(verbose: bool = False):
        """Register colormaps with matplotlib.

        Args:
            verbose (bool): If True, print information to command line
        """
        for g in SkaisColorMaps.aslist:
            if verbose:
                print(g.name)
            plt.register_cmap(name=g.name, cmap=g)


if __name__ == "__main__":
    SkaisColorMaps.plot_gradients()
    plt.show()



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Utilities for handling cosmology-dependent calculations."""

from dataclasses import dataclass, field
from functools import partial

import numpy as np
from scipy.integrate import solve_ivp
from astropy import units as au
from astropy import constants as ac
from typing import TypeVar

TCosmo = TypeVar("TCosmo", bound="CosmoModel")


@dataclass
class CosmoModel:
    """Set cosmological parameter for distance calculations, mass projections, etc.

    Args:
        omega_m (float): matter energy fraction
        omega_l (float): dark energy fraction
        omega_k (float): curvature fraction
        omega_r (float): radiation energy fraction (will be calculated
                          assuming flat universe if not given)
        h (float): 'little' H (H0 / 100 km/s/Mpc)
        c (ac.Constant): speed of light
        z (float): redshift
    """

    omega_m: float = 0.279952
    omega_l: float = 0.72
    omega_k: float = 0.0
    # omega_r: float = field(init=False, default=None, hash=False)
    _omega_r: float | None = field(init=False, default=None, hash=False)
    z: float = 0.0
    c: ac.Constant = field(default_factory=lambda: ac.c)
    h: float = 0.718
    u_H0: au.CompositeUnit = au.km / au.s / au.Mpc

    def __str__(self) -> str:
        """Instance string representation."""
        return (
            f"<CosmoModel[Ω_m:{self.omega_m:2.6f} Ω_l:{self.omega_l:2.6f} "
            f"Ω_k:{self.omega_k:2.6f} Ω_r:{self.omega_r:2.6f} "
            f"h:{self.h:2.4f} c:{self.c.to(au.km/au.s)} z:{self.z}]>"
        )

    def __repr__(self) -> str:
        """Instance string representation."""
        return self.__str__()

    @property
    def omega_r(self) -> float:
        """The radiation density parameter getter."""
        if self._omega_r is None:
            self._omega_r = (1 + self.omega_k) - self.omega_m - self.omega_l
        return self._omega_r

    @omega_r.setter
    def omega_r(self, r: float | None):
        """The radiation density parameter setter."""
        if r is None:
            r = (1 + self.omega_k) - self.omega_m - self.omega_l
        if isinstance(r, property):
            r = CosmoModel._omega_r
        self._omega_r = r

    @property
    def a(self) -> float:
        """The scale parameter getter."""
        if self.z:
            return 1.0 / (1 + self.z)
        return 1.0  # default: z = 0

    @a.setter
    def a(self, a: float):
        """The scale parameter setter."""
        self.z = (1.0 / a) - 1

    @property
    def rho_crit(self) -> au.Quantity:
        """The critical density getter."""
        H0 = self.H0
        return (3 * H0**2 / (8 * np.pi * ac.G)).to(au.Msun * au.Mpc ** (-3))

    @property
    def H0(self) -> au.Quantity:
        """The Hubble constant getter."""
        return self.h * 1e2 * self.u_H0

    @H0.setter
    def H0(self, H0: au.Quantity):
        """The Hubble constant setter."""
        self.h = H0 / (1e2 * self.u_H0)

    def H(self, a: float | None = None) -> au.Quantity:
        """The Hubble parameter at a given scale factor `a`.

        Args:
            a: scale factor

        Returns:
            (float): value of the Hubble parameter
        """
        if a is None and self.z and self.a:
            a = self.a
        elif a is None:
            a = 1.0
        if self.omega_r is None:
            self.omega_r
        return (
            self.H0
            * (self.omega_m / a**3 + self.omega_r / a**4 + self.omega_k / a**2 + self.omega_l)
            ** 0.5
        )

    @staticmethod
    def d_comov(a: float, r: float, cosmo_model: TCosmo | None = None) -> au.Quantity:
        """Calculate the comoving distance from scale factor (for solve_ivp).

        Args:
            a: scale factor
            r: distance
            cosmo_model: cosmological parameter lookup class

        Returns:
            (float): comoving radial distance (scale-free)
        """
        if cosmo_model is None:
            cosmo_model = CosmoModel()
        return 1.0 / (a * a * cosmo_model.H(a))

    @staticmethod
    def d_z(
        z: float | None = None,
        cosmo_model: TCosmo | None = None,
        scaled: bool = False,
    ) -> au.Quantity:
        """Angular distance d_z from a redshift z within given cosmology.

        Args:
            z: redshift
            cosmo_model: cosmological parameter lookup class
            scaled: return result with c/H0 in units of kpc

        Return:
            (float): apparent distance d_z
        """
        if cosmo_model is None:
            cosmo_model = CosmoModel()
        comov = partial(CosmoModel.d_comov, cosmo_model=cosmo_model)
        if z is None and cosmo_model.a:
            a = cosmo_model.a
        elif z is None:
            a = 0
        else:
            a = 1.0 / (1.0 + z)
        a_lim = [a, 1]
        res = solve_ivp(comov, a_lim, [0])
        if res["success"]:
            r = res["y"][0]
            D = a * (r[-1] - r[0])
            if scaled:
                return CosmoModel.d_z2kpc(D)
            return D
        raise ValueError(f"No solution found for the inputs z={z}, " f"cosmo_model={cosmo_model}")

    @staticmethod
    def d_z2kpc(
        distance: float | np.ndarray, cosmo_model: TCosmo | None = None
    ) -> au.Quantity:
        """Given scale-less distance d, return scaled distance c/H0 * d [kpc].

        Args:
            distance: scale-less distance
            cosmo_model: cosmological parameter lookup class
        """
        if cosmo_model is None:
            cosmo_model = CosmoModel()
        c = cosmo_model.c
        H0 = cosmo_model.H0
        return (distance * c / H0).to(au.kpc)

    @staticmethod
    def arcsec2kpc(
        z: float,
        dist_z: float | None = None,
        cosmo_model: TCosmo | None = None,
    ) -> au.Quantity:
        """Angular distance d_z from a redshift z within given cosmology.

        Args:
            z: redshift
            dist_z: scaled comoving distance
            cosmo_model: cosmological parameter lookup class

        Return:
            (float): kpc/arcsec scaling within given redshift and cosmology
        """
        if dist_z is None:
            dist_z = CosmoModel.d_z(z, cosmo_model=cosmo_model, scaled=True)
        d_kpc_arcsec = (dist_z / au.rad).to(au.kpc / au.arcsec)
        return d_kpc_arcsec


if __name__ == "__main__":
    import matplotlib.pyplot as plt

    # distance calculation test z=3
    cosmo = CosmoModel(z=3)
    dz = CosmoModel.d_z(cosmo.z, cosmo)
    dang = CosmoModel.d_z2kpc(dz, cosmo)
    dcomov = (1 + cosmo.z) * dang
    a2k = CosmoModel.arcsec2kpc(cosmo.z, dang, cosmo)
    print(f"Assumed cosmology:\t{cosmo}")
    print(f"Scale @ z={cosmo.z}:         \t{a2k}")
    print(f"Angular dist. @ z={cosmo.z}: \t{dang.to(au.Mpc)}")
    print(f"Comoving dist. @ z={cosmo.z}:\t{dcomov.to(au.Mpc)}")
    # distance calculation test z=1
    cosmo = CosmoModel(z=1)
    dz = CosmoModel.d_z(cosmo.z, cosmo)
    dang = CosmoModel.d_z2kpc(dz, cosmo)
    dcomov = (1 + cosmo.z) * dang
    a2k = CosmoModel.arcsec2kpc(cosmo.z, dang, cosmo)
    print(f"Assumed cosmology:\t{cosmo}")
    print(f"Scale @ z={cosmo.z}:         \t{a2k}")
    print(f"Angular dist. @ z={cosmo.z}: \t{dang.to(au.Mpc)}")
    print(f"Comoving dist. @ z={cosmo.z}:\t{dcomov.to(au.Mpc)}")
    # distance calculation test z=0
    cosmo = CosmoModel(z=0.01)
    dz = CosmoModel.d_z(cosmo.z, cosmo)
    dang = CosmoModel.d_z2kpc(dz, cosmo)
    dcomov = (1 + cosmo.z) * dang
    a2k = CosmoModel.arcsec2kpc(cosmo.z, dang, cosmo)
    print(f"Assumed cosmology:\t{cosmo}")
    print(f"Scale @ z={cosmo.z}:         \t{a2k}")
    print(f"Angular dist. @ z={cosmo.z}: \t{dang.to(au.Mpc)}")
    print(f"Comoving dist. @ z={cosmo.z}:\t{dcomov.to(au.Mpc)}")

    # plotting angular distances [0.2-13]
    cosmo = CosmoModel()
    redshifts = np.linspace(0.2, 13, 300)
    d_zs = np.array([CosmoModel.d_z(z, cosmo) for z in redshifts])
    dangs = CosmoModel.d_z2kpc(d_zs, cosmo)
    plt.figure(figsize=(16, 9), dpi=100)
    plt.plot(redshifts, dangs.to(au.Mpc))
    plt.xlabel("redshifts")
    plt.ylabel("D$_{ang}$ [Mpc]")
    plt.show()



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Package metadata."""

__version__ = "0.1.2"



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Generate maps from simulations."""

from pathlib import Path
import gc
import datetime
import csv
import logging
import hydra
from hydra.core.hydra_config import HydraConfig
from hydra.utils import instantiate
from omegaconf import OmegaConf, DictConfig
import numpy as np
from matplotlib.colors import Colormap
from astropy import units as au

# from astropy.visualization import ImageNormalize, MinMaxInterval
# from astropy.visualization import AsinhStretch
# from astropy.visualization import LogStretch
from typing import Any
from collections.abc import Iterable
import skais_mapper
from skais_mapper.utils import get_run_id, compress_encode
from skais_mapper.data import Img2H5Buffer
from skais_mapper.raytrace import voronoi_NGP_2D
from skais_mapper.simobjects import TNGGalaxy
from skais_mapper.plotting import plot_image


def map_TNG_sample(
    obj: TNGGalaxy,
    gid: int,
    group: str = "gas",
    projected_unit: au.Unit = None,
    cmap: Colormap | None = None,
    hdf5_file: str | Path | None = None,
    hdf5_save: bool = True,
    npy_save: bool = False,
    png_save: bool = False,
    subdir_save: bool = False,
    grid_size: int = 512,
    fh: float = 3,
    rot: list[float] | tuple[float, float] | None = None,
    xaxis: int = 0,
    yaxis: int = 1,
    periodic: bool = False,
    rng_seed: int = 42,
    flag_lim: float = 0,
    flag_N: int = 64,
    dry_run: bool = False,
    verbose: bool = True,
):
    """Project a subfind ID from an IllstrisTNG snapshot.

    Args:
        obj: Instance at a set snapshot, pointing at set subfind ID.
        gid: Galaxy/halo index.
        group: Galaxy property of the map, one of [gas,star,gas,hi,hi/21cm,temp,bfield].
        projected_unit: Units in which the map is to be projected.
        cmap: Colormap for map plot.
        hdf5_file: Basename of the HDF5 file.
        hdf5_save: If True, save map to HDF5 file.
        npy_save: If True, save map as numpy binary files.
        png_save: If True, save map plot as PNG file.
        subdir_save: If True, saves numpy binary and PNG files in corresponding subdirectories.
        grid_size: The size of the maps/images. Default: 512.
        fh: Expansion factor for the SPH particle radii.
        rot: Angles by which the sample is rotated before projection.
        xaxis: Projection axis for x.
        yaxis: Projection axis for y.
        periodic: Use periodic boundary conditions for the projection (for metadata).
        rng_seed: Seed for the random number generation.
        flag_lim: Flag the map in the metadata if N pixel values fall below the limit.
        flag_N: The number of pixels before an image is flagged.
        dry_run: If True, nothing is saved and expensive computation is skipped.
        verbose: If True, print status updates to command line.
    """
    # gather settings
    kwargs: dict[str, Any] = {
        "use_half_mass_rad": True,
        "fh": fh,
        "grid_size": grid_size,
        "xaxis": xaxis,
        "yaxis": yaxis,
        "periodic": periodic,
        "rot": rot,
        "verbose": verbose,
    }
    post_hook = None
    # set up configs for group
    if group == "gas":
        kwargs["keys"] = ["particle_positions", "masses", "radii", "center"]
        if projected_unit is None:
            projected_unit = au.Msun / au.kpc**2
    elif group == "hi":
        kwargs["keys"] = ["particle_positions", "m_HI", "radii", "center"]
        if projected_unit is None:
            projected_unit = au.Msun / au.kpc**2
    elif group == "hi/21cm":
        kwargs["keys"] = ["particle_positions", "m_HI", "radii", "center"]
        kwargs["assignment_func"] = voronoi_NGP_2D
        kwargs["tracers"] = 128
        kwargs["divisions"] = 2
        pixel_size = 1.0 / grid_size
        z, h, H0, Hz = (
            obj.cosmology.z,
            obj.cosmology.h,
            obj.cosmology.H0,
            obj.cosmology.H(obj.cosmology.a),
        )
        sigma_crit = obj.cosmology.rho_crit

        def post_hook(x, y):
            return (
                189
                * au.mK
                * h
                * (1 + z) ** 2
                * (H0 / Hz)
                * x
                / ((y[1] - y[0]) * pixel_size * sigma_crit)
            )

        if projected_unit is None:
            projected_unit = au.mK
        flag_lim, flag_N = 0, int(grid_size**2 / 10)
    elif group == "temp":
        kwargs["keys"] = [
            "particle_positions",
            ("masses", "temperature"),
            "radii",
            "center",
        ]
        if projected_unit is None:
            projected_unit = au.K
    elif group == "bfield":
        kwargs["keys"] = [
            "particle_positions",
            ("masses", "magnetic_field_strength"),
            "radii",
            "center",
        ]
        if projected_unit is None:
            projected_unit = au.Gauss
    elif group == "star":
        kwargs["keys"] = ["particle_positions", "masses", "radii", "center"]
        if projected_unit is None:
            projected_unit = au.Msun / au.kpc**2
    elif group == "dm":
        kwargs["keys"] = ["particle_positions", "masses", "radii", "center"]
        if projected_unit is None:
            projected_unit = au.Msun / au.kpc**2
    if isinstance(kwargs["keys"][1], tuple | list):
        keys = kwargs.pop("keys")
        quantity, extent, N = obj.generate_map(keys=keys, **kwargs)
        keys[1] = keys[1][0]
        weight_map, _, _ = obj.generate_map(keys=keys, **kwargs)
        projected = np.zeros_like(quantity.value)
        np.place(
            projected,
            weight_map.value != 0,
            quantity.value[weight_map.value != 0] / weight_map.value[weight_map.value != 0],
        )
        projected *= quantity.unit / weight_map.unit
    else:
        # allocate arrays and raytrace
        projected, extent, N = obj.generate_map(**kwargs)
    if post_hook is not None:
        projected = post_hook(projected, extent)
    # convert to chosen units
    projected = projected.to(projected_unit)
    # check for potential problems
    flag = 0
    if np.sum(projected.value < flag_lim) > flag_N:
        print("Potential issue with projection, flagging image...")
        flag = 1
    has_bh = obj.N_particles_type[-1]
    # plot data
    rot_str = f"_rotxy.{rot[0]}.{rot[1]}" if rot is not None else ""
    bname = f"{str(Path(group).stem)}_tng50-1.{obj.snapshot:02d}.gid.{obj.halo_index:07d}{rot_str}"
    md = {
        "class": group,
        "gid": obj.halo_index,
        "snapshot": obj.snapshot,
        "units": f"{projected.unit}",
        "extent": extent.value,
        "units_extent": f"{extent.unit}",
        "name": bname,
        "num_particles": N,
        "rotxy": rot if rot is not None else (0, 0),
        "N_particle_flag": flag,
        "has_bh": has_bh,
        "rng_seed": rng_seed,
    }
    plot_image(
        projected,
        info=md,
        cbar=True,
        norm="log" if group != "hi/21cm" else None,
        savefig=png_save and not dry_run,
        path=hdf5_file.parent / group / "png" / f"{bname}.png"
        if subdir_save
        else hdf5_file.parent / f"{bname}.png",
        show=dry_run,
        close=not dry_run,
        verbose=verbose
    )
    # save data
    if npy_save:
        npname = f"{bname}.units.{projected.unit}.extent.{extent[1] - extent[0]:4.8f}.npy".replace(
            " ", ""
        ).replace("/", "_")
        np_dir = hdf5_file.parent
        if subdir_save:
            np_dir = np_dir / group / "npy"
        if not dry_run:
            if not np_dir.exists():
                np_dir.mkdir(parents=True)
            np.save(np_dir / npname, projected.value)
        if verbose:
            print(f"Saving to [npy]: {np_dir / npname}")
    if hdf5_save and hdf5_file is not None:
        if subdir_save:
            hdf5_file = hdf5_file.parent / group / "hdf5" / hdf5_file.name
            if not hdf5_file.parent.exists() and not dry_run:
                hdf5_file.parent.mkdir(parents=True)
        img2h5 = Img2H5Buffer(target=hdf5_file, size="2G")
        img_target = f"{str(hdf5_file)}"
        mdt_target = f"{str(hdf5_file)}"
        img_h5group = f"{group}/images"
        mdt_h5group = f"{group}/metadata/"
        if not dry_run:
            img_h5group = f"{group}/images"
            img2h5.inc_write(img_target, data=projected.value, group=img_h5group, verbose=verbose)
            mdt_h5group = f"{group}/metadata/{img2h5.index:04d}"
            img2h5.inc_write(mdt_target, data=md, group=mdt_h5group, verbose=verbose)
        if verbose:
            print(f"Saving to [hdf5]: {img_target}:{img_h5group}")
            print(f"Saving to [hdf5]: {mdt_target}:{mdt_h5group}")


def map_TNG_galaxies(
    snapshots: list[int],
    gids: int | Iterable[int],
    groups: list[str] | None = None,
    output: str | None = None,
    src_dir: str | None = None,
    sim_type: str = "illustris/tng50-1",
    part_max: int | None = None,
    part_min: int | None = 20_000,
    retries: int | None = None,
    subfind_limit: int | None = 15_000,
    grid_size: int = 512,
    rotations: np.ndarray | None = None,
    random_rotations: bool = True,
    rng_seed: int = 42,
    dry_run: bool = False,
    verbose: bool = True,
):
    """Generate any number of maps from an IllustrisTNG snapshot(s).

    Args:
        snapshots: Snapshots number of the IllustrisTNG run.
        gids: Subfind IDs, i.e. galaxies, from which to generate maps.
        groups: Galaxy properties to map, e.g. star, gas, or dm.
        output: Output filename. Can have format fields '{}' for group and snapshot.
        src_dir: Path to the root of the simulation snapshots.
        sim_type: Simulation type (should correspond to the subpath in `src_dir`).
        part_max: Maximum number of particles to use for map generation.
        part_min: Minimum number of particles to use for map generation.
        retries: If not None, sets the maximum number of replacement candidates for skipped groups.
        subfind_limit: If not None, sets the maximum subfind ID allowed as replacement.
        grid_size: The size of the maps/images. Default: 512.
        rotations: List of angle pairs (theta, phi) per rotation for each subfind ID;
          e.g. for 4 separate rotations per subfind ID, its shape is (len(gids), 4, 2).
        random_rotations: If True, use random rotations (2 per subfind ID) to
          augment the dataset.
        rng_seed: Random number seed.
        dry_run: If True, nothing is saved and expensive computation is skipped.
        verbose: If True, print status updates to command line.
    """
    snapshots = list(snapshots)
    gids = list(range(gids)) if isinstance(gids, int) else list(gids)
    # list of groups to generate
    if groups is None:
        groups = ["gas"]
    Ng = len(gids) * len(groups)
    skip_count = [0] * (Ng // len(groups))
    # gather paths
    src_path = Path(src_dir) if src_dir is not None else Path("./simulations")
    tng_path = src_path / sim_type
    if output is None:
        hdf5_file = Path(
            str(datetime.datetime.now().date()).replace("-", "")
            + f"_{tng_path.name}.{{}}.2D.{{}}.hdf5"
        )
    else:
        hdf5_file = Path(output)
    csv_file = hdf5_file.parent / f"{get_run_id()}.group_particles.csv"
    if not csv_file.exists() and not dry_run:
        csv_file.parent.mkdir(parents=True, exist_ok=True)
        with csv_file.open(mode="w", newline="") as fcsv:
            writer = csv.writer(fcsv)
            writer.writerow(
                [
                    "snapshot",
                    "gid",
                    "N_particles_gas",
                    "N_particles_dm",
                    "N_particles_stars",
                    "N_particles_bh",
                ]
            )

    # Resolve paths and verify existance
    print("Resolved paths:")
    print("Source path:", src_path.resolve())
    print(f"{src_path.resolve()}: exists", src_path.exists())
    print("tng50-1:    ", tng_path.resolve())
    print(f"{tng_path.resolve()}: exists", tng_path.exists())
    print("Output:     ", hdf5_file)

    # Precompute all rotations
    if random_rotations:
        rng = np.random.default_rng(rng_seed)
        N_rot = Ng
        rotations = np.stack(
            (
                rng.integers(25, 180, size=N_rot),
                rng.integers(25, 90, size=N_rot),
                rng.integers(90, 270, size=N_rot),
            )
        )
        rotations = np.vstack(
            (
                rotations,
                rotations[0] + rng.integers(20, 40, size=N_rot),
                rotations[1] + rng.integers(70, 110, size=N_rot),
                rotations[2] + rng.integers(40, 90, size=N_rot),
            )
        ).T.reshape(Ng, 3, 2)

    # Generation loop
    for snap_id in snapshots:
        # run through galaxies
        for i, gid in enumerate(gids):
            if Ng < 0:
                break
            angles = [] if rotations is None else rotations[i]
            for j, group in enumerate(groups):
                p_group = group if group in ["gas", "star", "dm"] else "gas"
                tng_src = TNGGalaxy(
                    tng_path,
                    snap_id,
                    halo_index=gids[0],
                    particle_type=p_group,
                    as_float32=True,
                    verbose=True,
                )
                if gid != tng_src.halo_index:
                    tng_src.halo_index = gid
                if p_group != tng_src.particle_type:
                    tng_src.particle_type = group
                print(f"\n# Snapshot {snap_id}, subhalo {gid}, {group}")
                # check if number of particles in halo is within accepted range
                if (part_max is not None and part_max < tng_src.N_particles_type[0]) or (
                    part_min is not None and tng_src.N_particles_type[0] < part_min
                ):
                    if verbose:
                        print(
                            "Skipping candidate due to low particle number"
                            f" {tng_src.N_particles_type[0]}..."
                        )
                    # add another group candidate below the limit
                    if retries is not None and skip_count[-1] >= retries:
                        Ng -= 1
                    elif subfind_limit is not None:
                        while gid in gids and gid <= subfind_limit:
                            gid += 1
                        gids.append(gid)
                        if rotations is not None:
                            rotations = np.concatenate(
                                (rotations, rotations[i][np.newaxis, ...]), axis=0
                            )
                        skip_count.append(skip_count[-1] + 1)
                    break
                Ng -= 1

                # construct actual hdf5 filename
                if str(hdf5_file).count("{") == 2:
                    out_hdf5 = Path(str(hdf5_file).format(snap_id, group.replace("/", ".")))
                elif str(hdf5_file).count("{") == 1:
                    out_hdf5 = Path(str(hdf5_file).format(group.replace("/", ".")))
                else:
                    out_hdf5 = hdf5_file
                # generate maps, plots, and save to files
                map_TNG_sample(
                    tng_src,
                    gid,
                    group=group,
                    hdf5_file=out_hdf5,
                    grid_size=grid_size,
                    fh=3 if group == "dm" else 2 if group == "star" else 1,
                    rng_seed=rng_seed,
                    rot=None,
                    hdf5_save=True,
                    npy_save=True,
                    png_save=True,
                    subdir_save=True,
                    dry_run=dry_run,
                    verbose=verbose,
                )
                for theta, phi in angles:
                    map_TNG_sample(
                        tng_src,
                        gid,
                        group=group,
                        hdf5_file=out_hdf5,
                        grid_size=grid_size,
                        fh=3 if group == "dm" else 2 if group == "star" else 1,
                        rng_seed=rng_seed,
                        rot=(theta, phi),
                        hdf5_save=True,
                        npy_save=True,
                        png_save=True,
                        subdir_save=True,
                        dry_run=dry_run,
                        verbose=verbose,
                    )
            if csv_file.exists():
                with open(csv_file, "a", newline="") as fcsv:
                    writer = csv.writer(fcsv)
                    writer.writerow(
                        [snap_id]
                        + [gid]
                        + tng_src.N_particles_type[0:2]
                        + tng_src.N_particles_type[4:],
                    )
            if verbose:
                print(
                    f"Number of particles in group: {tng_src.N_particles_type[0]} [gas]"
                    f" | {tng_src.N_particles_type[tng_src.p_idx]} [{group}]"
                )
            gc.collect()


@hydra.main(config_path="configs", config_name="config", version_base=None)
def run(cfg: DictConfig | dict):
    """Main routine for generating any number of maps from a simulation snapshot(s).."""
    log = logging.getLogger(__name__)
    output_dir = HydraConfig.get().runtime.output_dir
    opt = instantiate(cfg, _convert_="all")
    if not cfg.exclude_git_state:
        log.info(f"Git state: {skais_mapper.GIT_STATE}")
    if cfg.include_git_diff:
        for d in skais_mapper.GIT_DIFF:
            log.info(f"Git diff: {compress_encode(d)}")
    log.info(f"Job id: {opt['run_id']}")
    log.info(f"Output directory: {output_dir}")
    if cfg.verbose:
        print("Configuration:")
        print(OmegaConf.to_yaml(cfg))

    if cfg.save_configs:
        hydra_subdir = Path(HydraConfig.get().output_subdir)
        src_file = hydra_subdir / "config.yaml"
        dst_file = Path(opt["output"])
        if dst_file.suffix not in [".yaml", ".yml"]:
            dst_file = Path(f"./{opt['run_id']}.yaml")
        dst_file.write_bytes(src_file.read_bytes())

    if "tng" in opt["simulation_type"].lower():
        np.random.seed(opt["random_seed"])
        map_TNG_galaxies(
            opt["snapshots"],
            range(*opt["num_samples"]),
            groups=opt["groups"],
            sim_type=opt["simulation_type"],
            output=opt["output"],
            src_dir=opt["source"],
            part_min=opt["part_min"],
            retries=opt["retries"],
            subfind_limit=opt["subfind_limit"],
            rng_seed=opt["random_seed"],
            grid_size=opt["grid_size"],
            dry_run=opt["dry_run"],
            verbose=opt["verbose"],
        )



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Inspect and create configuration files."""

import logging
from pathlib import Path
import hydra
from hydra.core.hydra_config import HydraConfig
from hydra.utils import instantiate
from omegaconf import OmegaConf, DictConfig
import skais_mapper
from skais_mapper.utils import compress_encode


@hydra.main(config_path="configs", config_name="config", version_base=None)
def create(cfg: DictConfig | dict):
    """Main configuration creation routine."""
    log = logging.getLogger(__name__)
    output_dir = HydraConfig.get().runtime.output_dir
    opt = instantiate(cfg)
    if not cfg.exclude_git_state:
        log.info(f"Git state: {skais_mapper.GIT_STATE}")
    if cfg.include_git_diff:
        for d in skais_mapper.GIT_DIFF:
            log.info(f"Git diff: {compress_encode(d)}")
    log.info(f"Job id: {opt.run_id}")
    log.info(f"Output directory: {output_dir}")
    if opt.verbose:
        print("Configuration:")
        print(OmegaConf.to_yaml(cfg))

    if opt.save_configs:
        hydra_subdir = Path(HydraConfig.get().output_subdir)
        src_file = hydra_subdir / "config.yaml"
        dst_file = Path(opt.output)
        if dst_file.suffix not in [".yaml", ".yml"]:
            dst_file = Path(f"./{opt.run_id}.yaml")
        dst_file.write_bytes(src_file.read_bytes())



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Illustris file i/o for FoF and Subfind group catalog.

Adapted from: https://github.com/illustristng/illustris_python
"""

import os
import numpy as np
from tqdm import trange
import h5py
from numpy.typing import NDArray
from skais_mapper.illustris.util import IllustrisH5File


def get_path(base_path: str, snapshot: int, partition: int = 0) -> str:
    """Get absolute path to a group catalog HDF5 file (modify as needed).

    Args:
        base_path: Base path to the Illustris(TNG) snapshots
        snapshot: Snapshot ID {0-99}
        partition: Subfile partition ID {0-600+}

    Returns:
        (str): Absolute path to a group catalog HDF5 file
    """
    gc_dir = os.path.join(base_path, f"groupcats/{snapshot:03d}")
    filepath = os.path.join(gc_dir, f"groups_{snapshot:03d}.{partition:d}.hdf5")
    filepath_alt = filepath.replace("groups_", "fof_subhalo_tab_")
    if os.path.isfile(filepath):
        return filepath
    return filepath_alt


def get_offset_path(base_path: str, snapshot: int, *args) -> str:
    """Get absolute path to a separate offset file (modify as needed).

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        *args: Dummy arguments for compatibility.

    Returns:
        (str): Absolute path to a group catalog's offsets HDF5 file.
    """
    basename = f"offsets/offsets_{snapshot:03d}.hdf5"
    offset_path = os.path.join(base_path, basename)
    return offset_path


def load_catalog(
    base_path: str,
    snapshot: int,
    key: str,
    key_ref: str,
    fields: list = None,
    as_float32: bool = False,
    as_array: bool = True,
    with_pbar: bool = True,
) -> dict | NDArray:
    """Load either halo or subhalo information from the group catalog.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        key: Group name from the HDF5 group catalog, e.g. 'Group' or 'Subhalo'.
        key_ref: Group name reference string in the HDF5 group catalog's
          header keys, e.g. 'groups' or 'subgroups'/'subhalos'
        fields: Fields to be loaded for the corresponding group,
          e.g. ['GroupPos', 'GroupMass'] or 'SubhaloGasMetalFractions'.
        as_float32: Load float64 data types as float32 (to save memory).
        as_array: Return a numpy array instead of a dictionary; takes
          effect only if a single field was requested.
        with_pbar: If True, a progress bar will show the current status.

    Returns:
        (dict | numpy.ndarray): A dictionary of the loaded data
    """
    data = {}
    if fields is None:
        fields = []
    elif isinstance(fields, str | bytes):
        fields = [fields]
    # load header from first partition
    IllustrisH5File.path_func = get_path
    header = load_header(base_path, snapshot)
    group = load_group(base_path, snapshot, key)
    if f"N{key_ref}_Total" not in header and key_ref == "subgroups":
        key_ref = "subhalos"
    data["count"] = header.get(f"N{key_ref}_Total", None)
    if not data["count"]:
        print(f"Warning: zero groups, empty return (snap='{snapshot}').")
        return data
    if not fields:
        fields = list(group.keys())
    for field in fields:
        if field not in group.keys():
            raise KeyError(f"Group catalog does not have requested field [{field}]!")
        # replace local length with global
        shape = list(group[field].shape)
        shape[0] = data["count"]
        dtype = group[field].dtype
        if dtype == np.float64 and as_float32:
            dtype = np.float32
        # allocate data arrays
        data[field] = np.zeros(shape, dtype=dtype)
    group._id.close()
    # loop over partitions
    arr_offset = 0
    if with_pbar:
        partition_iterator = trange(header["NumFiles"])
    else:
        partition_iterator = range(header["NumFiles"])
    for i in partition_iterator:
        f = IllustrisH5File(base_path, snapshot, i)
        # if partition is empty
        if not f["Header"].attrs[f"N{key_ref}_ThisFile"]:
            continue
        # loop over each field
        for field in fields:
            if field not in f[key].keys():
                raise KeyError(f"Group catalog does not have requested field [{field}]!")
            # shape and type
            shape = f[key][field].shape
            if len(shape) == 1:
                data[field][arr_offset : arr_offset + shape[0]] = f[key][field][0 : shape[0]]
            else:
                data[field][arr_offset : arr_offset + shape[0], :] = f[key][field][0 : shape[0], :]
        arr_offset += shape[0]
        f.close()
    if as_array and len(fields) == 1:
        return data[fields[0]]
    return data


def load_subhalos(base_path: str, snapshot: int, **kwargs) -> dict | NDArray:
    """Load all subhalo information from the entire group catalog for one snapshot.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        **kwargs: Additional keywords such as
          fields: Fields to be loaded for the corresponding group,
            e.g. ['GroupPos', 'GroupMass'] or 'SubhaloGasMetalFractions'.
          as_float32: Load float64 data types as float32 (to save memory).
          as_array: Return a numpy array instead of a dictionary; takes
            effect only if a single field was requested.
          with_pbar: If True, a progress bar will show the current status.

    Returns:
        (dict | numpy.ndarray): A dictionary of the loaded data
    """
    return load_catalog(base_path, snapshot, "Subhalo", "subgroups", **kwargs)


def load_halos(base_path: str, snapshot: int, **kwargs) -> dict | NDArray:
    """Load all halo information from the entire group catalog for one snapshot.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        fields: Fields to be loaded for the corresponding group,
          e.g. ['GroupPos', 'GroupMass'] or 'SubhaloGasMetalFractions'.
        as_float32: Load float64 data types as float32 (to save memory).
        as_array: Return a numpy array instead of a dictionary; takes
          effect only if a single field was requested.
        with_pbar: If True, a progress bar will show the current status.
        **kwargs: Additonal keywords for `load_catalog`

    Returns:
        (dict | numpy.ndarray): A dictionary of the loaded data
    """
    return load_catalog(base_path, snapshot, "Group", "groups", **kwargs)


def load_header(
    base_path: str,
    snapshot: int,
    as_dict: bool = True,
) -> dict | h5py.Group:
    """Load the header of a group catalog.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        as_dict: If True, a dictionary is returned, otherwise as
          h5py.Group object.

    Returns:
        (dict): The header of the group catalog HDF5 file as a dictionary
    """
    with IllustrisH5File(base_path, snapshot, path_func=get_path) as f:
        if as_dict:
            header = dict(f["Header"].attrs.items())
        else:
            header = f["Header"]
    return header


def load_group(base_path: str, snapshot: int, key: str, as_dict: bool = False) -> dict | h5py.Group:
    """Load a specified HDF5 group from a group catalog.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        key: Group descriptor, i.e. a group from the HDF5 catalog
          e.g. 'Group' or 'Subhalo'.
        as_dict: If True, a dictionary is returned, otherwise as
          h5py.Group object.

    Returns:
        (dict): The HDF5 group from the group catalog HDF5 file as a dictionary

    Note: Remember to close the HDF5 file afterwards (use <group>._id.close()).
    """
    f = IllustrisH5File(base_path, snapshot, path_func=get_path)
    if key in f:
        if as_dict:
            group = dict(f[key])
        else:
            group = f[key]
    else:
        f.close()
        return None
    return group


def load(
    base_path,
    snapshot,
    subhalos_kwargs: dict = None,
    halos_kwargs: dict = None,
    header_kwargs: dict = None,
    **kwargs,
) -> dict:
    """Load complete group catalog all at once.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        subhalos_kwargs: Keyword arguments for loading subhalos.
        halos_kwargs: Keyword arguments for loading halos.
        header_kwargs: Keyword arguments for loading the header.
        **kwargs: Additional 

    Returns:
        (dict): A dictionary of the loaded data
    """
    data = kwargs
    if subhalos_kwargs is None:
        subhalos_kwargs = {}
    if halos_kwargs is None:
        halos_kwargs = {}
    if header_kwargs is None:
        header_kwargs = {}
    data["subhalos"] = load_subhalos(base_path, snapshot, **subhalos_kwargs)
    data["halos"] = load_halos(base_path, snapshot, **halos_kwargs)
    data["header"] = load_header(base_path, snapshot, **header_kwargs)
    return data


def load_single(base_path, snapshot, halo_id=-1, subhalo_id=-1) -> dict:
    """Fetch the complete group catalog information for a single halo or subhalo.

    Args:
        base_path: Base path to the Illustris(TNG) snapshots.
        snapshot: Snapshot ID {0-99}.
        halo_id: Group ID, i.e. halo ID value from the FOF catalog.
        subhalo_id: Group ID, i.e. subhalo ID value from the FOF catalog.

    Returns:
        (dict): A dictionary of the loaded data
    """
    if (halo_id < 0 and subhalo_id < 0) or (halo_id >= 0 and subhalo_id >= 0):
        raise ValueError("Must specify either halo_id or subhalo_id (not both).")
    key = "Subhalo" if subhalo_id >= 0 else "Group"
    group_id = subhalo_id if subhalo_id >= 0 else halo_id
    # old or new format
    if "fof_subhalo" in get_path(base_path, snapshot):
        # use separate 'offsets_nnn.hdf5' files
        with IllustrisH5File(base_path, snapshot, path_func=get_offset_path) as f:
            offsets = f["FileOffsets/" + key][()]
    else:
        # use header of group catalog
        with IllustrisH5File(base_path, snapshot, path_func=get_path) as f:
            offsets = f["Header"].attrs["FileOffsets_" + key]
    offsets = group_id - offsets
    file_id = np.max(np.where(offsets >= 0))
    group_offset = offsets[file_id]
    # load halo/subhalo fields into a dict
    data = {}
    with IllustrisH5File(base_path, snapshot, file_id, path_func=get_path) as f:
        for field in f[key].keys():
            data[field] = f[key][field][group_offset]
    return data


if __name__ == "__main__":
    # TODO: unittest
    IllustrisH5File.path_func = get_path
    tng_dir = "/scratch/data/illustris/tng50-1"
    snap_id = 99

    # test getting a file path
    filename_0 = get_path(tng_dir, snap_id)
    filename = get_path(tng_dir, snap_id, 10)
    print(filename_0)
    print(filename)

    # test opening a HDF5 file manually
    print("# Test opening a HDF5 file manually")
    with IllustrisH5File(tng_dir, snap_id) as f:
        print(f.keys())
        print(f["Header"])
        print(f["Config"])
        print(f["IDs"])
        print(f["Parameters"])
        print(f["Group"])
        print(f["Subhalo"])

    print("# Test loading header from HDF5 file")
    cat_header = load_header(tng_dir, snap_id, True)
    print(cat_header)

    # test loading main group from HDF5 file
    print("# Test loading main group from HDF5 file")
    group = load_group(tng_dir, snap_id, "Group")
    print(group)

    # test loading catalog from HDF5 file (w/ progress bar)
    print("# Test loading catalog from HDF5 file w/ progress bar")
    cat = load_catalog(tng_dir, snap_id, "Group", "groups", fields=["GroupPos"])
    print(cat)

    # test loading catalog from HDF5 file (w/o progress bar)
    print("# Test loading catalog from HDF5 file w/o progress bar")
    cat = load_catalog(tng_dir, snap_id, "Group", "groups", fields=["GroupPos"], with_pbar=False)
    print(cat)

    # test loading all data from HDF5 file
    print("# Test loading all data from HDF5 file")
    data = load(tng_dir, snap_id, with_pbar=True)



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Rotation operators for point cloud arrays."""

from typing import TypeVar
import numpy as np


TRotation = TypeVar("TRotation", bound="R")


class R:
    """Class of rotation operators."""

    def __init__(self, omega: np.ndarray | None = None):
        """Initialize rotator object.

        Args:
            omega: Rotation angle in
        """
        self.omega = omega

    def __call__(self, arr: np.ndarray, **kwargs) -> np.ndarray:
        """Rotate the input according to this rotation operator.

        Args:
            arr: The array to be rotated
            kwargs: Dummy keyword arguments

        Returns:
            arr: The rotated array
        """
        if self.omega is None:
            self.omega = self._x(0)
        arr = arr @ self.omega.transpose()
        return arr

    def __mul__(self, other: TRotation) -> TRotation:
        """Combine rotation operators through multiplication.

        Args:
            other: Another rotation object to be multiplied.

        Returns:
            obj: a new instance of R
        """
        return self.__class__(self.omega @ other.omega)

    @classmethod
    def x(cls, theta: float, degrees: bool = True) -> TRotation:
        """Rotation operator about the current x-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        omega = cls._x(theta, degrees=degrees)
        return cls(omega)

    @classmethod
    def y(cls, theta: float, degrees: bool = True) -> TRotation:
        """Rotation operator about the current y-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        omega = cls._y(theta, degrees=degrees)
        return cls(omega)

    @classmethod
    def z(cls, theta: float, degrees: bool = True) -> TRotation:
        """Rotation operator about the current z-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        omega = cls._z(theta, degrees=degrees)
        return cls(omega)

    @staticmethod
    def _x(theta: float, degrees: bool = True) -> np.ndarray:
        """Rotation matrix about the current x-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        if degrees:
            theta *= np.pi / 180.0
        return np.array(
            [
                [1, 0, 0],
                [0, np.cos(theta), -np.sin(theta)],
                [0, np.sin(theta), np.cos(theta)],
            ]
        )

    @staticmethod
    def _y(theta: float, degrees: bool = True) -> np.ndarray:
        """Rotation matrix about the current y-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        if degrees:
            theta *= np.pi / 180.0
        return np.array(
            [
                [np.cos(theta), 0, np.sin(theta)],
                [0, 1, 0],
                [-np.sin(theta), 0, np.cos(theta)],
            ]
        )

    @staticmethod
    def _z(theta: float, degrees: bool = True) -> np.ndarray:
        """Rotation matrix about the current z-axis by angle theta.

        Args:
            theta: Rotation angle in degrees or radians (see below)
            degrees: Rotation in degrees, if false in radians
        """
        if degrees:
            theta *= np.pi / 180.0
        return np.array(
            [
                [0, np.cos(theta), -np.sin(theta)],
                [0, np.sin(theta), np.cos(theta)],
                [0, 0, 1],
            ]
        )



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Tools for manipulating data objects from simulations."""

import time
from pathlib import Path
from typing import Any
from collections.abc import Callable
from numpy.typing import NDArray
import numpy as np
import scipy as sp
import astropy.units as au
import astropy.constants as ac
import skais_mapper.illustris as tng
from skais_mapper.cosmology import CosmoModel
from skais_mapper.rotations import R
from skais_mapper.raytrace import voronoi_RT_2D, voronoi_NGP_2D


__all__ = ["SPHGalaxy", "TNGGalaxy", "GasolineGalaxy"]


class SPHGalaxy:
    """A generic base SPH Galaxy simulation parser."""

    primary_hdf5_fields: dict[int, list[str]] = {0: [], 1: [], 4: []}
    optional_hdf5_fields: dict[int, list[str]] = {0: [], 1: [], 4: []}

    def __init__(
        self,
        ra: float = 0.0,
        dec: float = 0.0,
        distance: float = 3.0,
        peculiar_v: float = 0.0,
        rotation: Callable | None = None,
        cosmo_pars: dict | None = None,
        units: dict | None = None,
        particle_type: str | None = None,
        as_float32: bool = False,
        verbose: bool = False,
    ):
        """Initialize an SPHGalaxy base class instance.

        Args:
            ra: The right ascension sky coordinate
            dec: The declination sky coordinate
            distance: The distance from observer
            peculiar_v: The galaxy's peculiar velocity
            rotation: Arbitrary rotation of the galaxy
            cosmo_pars: Cosmology settings
            units: unit system in the simulations
            particle_type: the particle type name e.g. 'gas', 'dm', or 'star' (see illustris.util)
            as_float32: If True, load simulation data as 32-bit floats (for efficiency)
            verbose: If True, print information to the command line
        """
        self.as_float32 = as_float32
        self.header = self.load_header()
        cosmo_pars = self.load_cosmology(cosmo_pars)
        self.cosmology = CosmoModel(**cosmo_pars)
        if particle_type is None:
            particle_type = "gas"
        self._p_idx = tng.util.pidx_from_ptype(particle_type)
        self.data = self.load_data(as_float32=self.as_float32, verbose=verbose)
        self.ra = ra * au.deg
        self.dec = dec * au.deg
        self.distance = distance * au.Mpc
        self.peculiar_v = peculiar_v * au.km / au.s
        self.rotation = rotation
        if units:
            self.set_units(**units)
        self.verbose = verbose

    @property
    def p_idx(self) -> int:
        """Particle index {0: 'gas', 1: 'dm', 2: 'tracers', 3: 'stars', 4: 'BHs'}."""
        return self._p_idx

    @p_idx.setter
    def p_idx(self, p_idx: int | str):
        """Setter for particle index."""
        if isinstance(p_idx, int):
            self._p_idx = p_idx
        elif isinstance(p_idx, str):
            self._p_idx = tng.util.pidx_from_ptype(p_idx)
        else:
            self._p_idx = 0
        # reload data
        self.load_data(as_float32=self.as_float32, verbose=self.verbose)

    @property
    def particle_type(self) -> str:
        """Particle type {0: 'gas', 1: 'dm', 2: 'tracers', 3: 'stars', 4: 'BHs'}."""
        p_idx = self.p_idx
        return tng.util.ptype_from_pidx(p_idx)

    @particle_type.setter
    def particle_type(self, ptype: int | str):
        """Setter for particle type."""
        self.p_idx = ptype

    def load_header(self) -> dict:
        """Dummy method, to be overridden in subclasses.

        Returns:
            (dict): the header dictionary which should contain
                - a 'snapshot' and 'catalog' subdictionary
                - the 'snapshot' subdictionary should contain
                    - 'UnitLength_in_cm', 'UnitVelocity_in_cm_per_s', 'UnitMass_in_g'
                    - 'OmegaLambda', 'Omega0', 'Redshift' / 'Time', 'HubbleParam'
                    - optionally: 'OmegaK'
        """
        self.header = {}
        self.header["snapshot"] = {}
        self.header["snapshot"]["UnitLength_in_cm"] = 3.085678e21
        self.header["snapshot"]["UnitVelocity_in_cm_per_s"] = 1e5
        self.header["snapshot"]["UnitMass_in_g"] = 1.989e43
        return self.header

    def load_cosmology(self, cosmo_pars: dict | None, in_place: bool = True) -> dict:
        """Dummy method, to be overridden in subclasses.

        Returns:
            (dict): a dictionary with cosmological parameters pulled from a header
                or set manually. The dictionary will be used for the CosmoModel
                dataclass.
        """
        if cosmo_pars is None:
            cosmo_pars = {}
        if in_place:
            self.cosmology = CosmoModel(**cosmo_pars)
        return cosmo_pars

    def angular_distance(self, eps: float = 1e-4) -> float:
        """Calculate the angular distance of the simulation at the given redshift.

        Args:
            eps: minimum cutoff value for the redshift
        """
        z = self.cosmology.z
        z += eps if z < eps else 0
        dz = self.cosmology.d_z(z, cosmo_model=self.cosmology, scaled=False)
        dang = self.cosmology.d_z2kpc(dz, cosmo_model=self.cosmology)
        return dang

    def angular_resolution(self, eps: float = 1e-4) -> tuple[float, float]:
        """Calculate the angular scale of the simulation at the given redshift.

        Args:
            eps: minimum cutoff value for the redshift
        """
        z = self.cosmology.z
        z += eps if z < eps else 0
        dz = self.cosmology.d_z(z, cosmo_model=self.cosmology, scaled=True)
        arcsec2kpc = self.cosmology.arcsec2kpc(z, dz)
        return 1.0 / arcsec2kpc.to(au.kpc / au.deg), z

    @property
    def boxsize(self) -> au.Quantity:
        """The virtual boxsize of the simulation."""
        if not hasattr(self, "header"):
            self.header = self.load_header()
        L = self.units("l/h")
        box = self.header["snapshot"].get("BoxSize", 1) * L
        return box.to(au.kpc)

    def units(self, u: str) -> au.Quantity:
        """Get common units from descriptor strings ('l', 'm', 'v', and variants).

        Args:
            u: the unit string describing the dimensionality;
                l -> length;
                m -> mass;
                v -> velocity;
                t -> time;
                vp -> peculiar velocity;
                sqrtP -> square root of pressure;
                c[] -> comoving quantities;
                []/h -> quantities divided by the dimensionless Hubble const.

        Returns:
            (astropy.units.Quantity): specified (scaled/composite) unit.
        """
        h = self.cosmology.h
        a = self.cosmology.a
        match u:
            case "l":
                return self.UnitLength
            case "m":
                return self.UnitMass
            case "v":
                return self.UnitVelocity
            case "t":
                return self.UnitLength / self.UnitVelocity
            case "l/h":
                return self.UnitLength * (a / h)
            case "m/h":
                return self.UnitMass / h
            case "vp":
                return self.UnitVelocity * np.sqrt(a)
            case "sqrtP":
                return (
                    h
                    / a**2
                    * (self.UnitMass / self.UnitLength) ** (1.0 / 2)
                    / (self.UnitLength / self.UnitVelocity)
                )
            case _:
                return 1

    def set_units(self, length: float, velocity: float, mass: float):
        """Set units in the header attribute.

        Args:
            length: length unit
            velocity: velocity unit (sets the time unit implicitly)
            mass: mass unit
        """
        if not hasattr(self, "header"):
            self.header = self.load_header()
        self.header["snapshot"]["UnitLength_in_cm"] = length
        self.header["snapshot"]["UnitVelocity_in_cm_per_s"] = velocity
        self.header["snapshot"]["UnitMass_in_g"] = mass

    @property
    def UnitLength(self) -> au.Quantity:
        """The simulation's length unit as astropy.units.Quantity."""
        if not hasattr(self, "header"):
            self.header = self.load_header()
        h = self.header["snapshot"]
        if "UnitLength_in_cm" in h:
            length = h["UnitLength_in_cm"] * au.cm
        else:
            length = 3.085678e21 * au.cm
        if self.as_float32:
            return length.to(au.kpc).astype(np.float32)
        return length

    @property
    def UnitMass(self) -> au.Quantity:
        """The simulation's mass unit as astropy.units.Quantity."""
        if not hasattr(self, "header"):
            self.header = self.load_header()
        h = self.header["snapshot"]
        if "UnitMass_in_g" in h:
            mass = h["UnitMass_in_g"] * au.g
        else:
            mass = 1.989e43 * au.g
        if self.as_float32:
            return mass.to(au.Msun).astype(np.float32)
        return mass

    @property
    def UnitVelocity(self) -> au.Quantity:
        """The simulation's velocity unit as astropy.units.Quantity."""
        if not hasattr(self, "header"):
            self.header = self.load_header()
        h = self.header["snapshot"]
        if "UnitVelocity_in_cm_per_s" in h:
            velocity = h["UnitVelocity_in_cm_per_s"] * au.cm / au.s
        else:
            velocity = 1e5 * au.cm / au.s
        if self.as_float32:
            return velocity.astype(np.float32)
        return velocity

    def load_data(self, **kwargs) -> dict:
        """Dummy method, to be overridden in subclasses.

        Returns:
            (dict): the data dictionary loaded from the simulations containing:
                - 'Density', 'Masses', 'Coordinates', 'Velocities', 'InternalEnergy',
                  'ElectronAbundance'
                - optionally 'CenterOfMass', 'GFM_Metals' (axis 1 should contain
                  hydrogen gas fractions), 'NeutralHydrogenAbundance'
        """
        kwargs.setdefault(
            "fields",
            self.primary_hdf5_fields[self.p_idx] + self.optional_hdf5_fields[self.p_idx],
        )
        data = {k: None for k in kwargs["fields"]}
        return data

    @property
    def density(self) -> au.Quantity | None:
        """The simulation's density data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        M = self.units("m/h")
        perVol = np.power(self.units("l/h"), -3)
        if "Density" in self.data:
            return self.data["Density"][:] * M * perVol
        elif "SubfindDMDensity" in self.data:
            return self.data["SubfindDMDensity"][:] * M * perVol
        elif "SubfindDensity" in self.data:
            return self.data["SubfindDensity"][:] * M * perVol
        return None

    @property
    def masses(self) -> au.Quantity | None:
        """The simulation's mass data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        M = self.units("m/h")
        if "Masses" in self.data:
            return (self.data["Masses"][:] * M).to(au.Msun)
        elif self.particle_mass is not None:
            masses = np.ones(self.data["Coordinates"].shape[0])
            return masses * self.particle_mass
        return None

    @property
    def particle_mass(self) -> au.Quantity | float | None:
        """The simulation's particle mass (for constant mass particles)."""
        h = self.load_header()["snapshot"]
        M = self.units("m/h")
        p_mass = None
        if "MassTable" in h and self.p_idx in [1, 3]:
            p_mass = (h["MassTable"][self.p_idx] * M).to(au.Msun)
        return p_mass

    @property
    def particle_positions(self) -> au.Quantity | None:
        """The simulation's particle position data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        L = self.units("l/h")
        if "Coordinates" in self.data:
            return (self.data["Coordinates"][:] * L).to(au.kpc)
        return None

    @property
    def cell_positions(self) -> au.Quantity | None:
        """The simulation's cell position data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        L = self.units("l/h")
        if "CenterOfMass" in self.data:
            return (self.data["CenterOfMass"][:] * L).to(au.kpc)
        return None

    @property
    def velocities(self) -> au.Quantity | None:
        """The simulation's particle velocity data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        V = self.units("vp")
        if "Velocities" in self.data:
            return self.data["Velocities"][:] * V
        return None

    @property
    def internal_energy(self) -> au.Quantity | None:
        """The simulation's internal energy data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        U = np.power(self.units("v"), 2)
        if "InternalEnergy" in self.data:
            return self.data["InternalEnergy"][:] * U
        return None

    @property
    def x_e(self) -> au.Quantity | None:
        """The simulation's electron abundance data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        if "ElectronAbundance" in self.data:
            return self.data["ElectronAbundance"][:]
        return None

    @property
    def x_H(self) -> au.Quantity | None:
        """The simulation's neutral hydrogen abundance data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        if "GFM_Metals" in self.data:
            return self.data["GFM_Metals"][:, 0][:]
        return None

    @property
    def x_HI(self) -> au.Quantity | None:
        """The simulation's ionized hydrogen abundance data in corresponding units."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        if "NeutralHydrogenAbundance" in self.data:
            return self.data["NeutralHydrogenAbundance"][:]
        return None

    @property
    def n_H(self) -> au.Quantity | None:
        """The simulation's hydrogen number density data in corresponding units."""
        if self.density is None:
            return None
        x_H = self.x_H
        perVol = au.cm ** (-3)
        # choose constants in favorable units to avoid overflows
        m_p = ac.m_p.to(au.u)
        n_H = x_H * self.density / m_p
        return n_H.to(perVol)

    @property
    def m_H(self) -> au.Quantity | None:
        """The simulation's hydrogen mass data in corresponding units."""
        if self.masses is None:
            return None
        m_H = self.x_H * self.masses
        return m_H.to(au.Msun)

    @property
    def n_HI(self) -> au.Quantity | None:
        """The simulation's ionized hydrogen number density data in corresponding units."""
        if self.x_HI is None or self.n_H is None:
            return None
        return self.x_HI * self.n_H

    @property
    def m_HI(self) -> au.Quantity | None:
        """The simulation's ionized hydrogen mass data in corresponding units."""
        if self.x_HI is None or self.m_H is None:
            return None
        m_HI = self.x_HI * self.m_H
        return m_HI.to(au.Msun)

    def kd_tree(
        self,
        k: int = 8,
        threads: int = 1,
        epsilon: float = 1.0e-4,
        as_float32: bool = False,
        verbose: bool = True,
    ) -> au.Quantity | NDArray:
        """Compute the distance of each gas particle to its k nearest neighbors.

        Args:
            k: Number of nearest neighbors.
            threads: Number of computing threads.
            epsilon: Small increase of the box size to avoid segfaults
            as_float32: If True, load simulation data as 32-bit floats (for efficiency)
            verbose: If True, print out results to the command line.

        Returns:
            (np.ndarray): The distance of each particle to its farthest neighbor
        """
        p = self.particle_positions
        if p is None:
            return None
        p = p.to(au.kpc)
        boxsize = (1 + epsilon) * self.boxsize.to(au.kpc)
        if verbose:
            print("Building KDTree...")
        t_ini = time.time()
        kdtree = sp.spatial.cKDTree(p, leafsize=16, boxsize=boxsize)
        if verbose:
            print(f"Time to build KDTree: {time.time() - t_ini:.3f} secs.")
        if verbose:
            print("Querying KDTree...")
        dist, _ = kdtree.query(p, k, workers=threads)
        if verbose:
            print(f"Time to querying KDTree: {time.time() - t_ini:.3f} secs.")
        if as_float32:
            return (dist[:, -1] * au.kpc).astype(np.float32)
        return dist[:, -1] * au.kpc


class TNGGalaxy(SPHGalaxy):
    """IllustrisTNG Galaxy simulation parser."""

    primary_hdf5_fields: dict[int, list[str]] = {
        0: [
            "Coordinates",
            "Density",
            "Masses",
            "Velocities",
            "InternalEnergy",
            "ElectronAbundance",
            "MagneticField",
        ],
        1: ["Coordinates", "Potential", "SubfindDMDensity", "Velocities"],
        4: ["Coordinates", "Masses", "Velocities", "SubfindDensity"],
    }
    optional_hdf5_fields: dict[int, list[str]] = {
        0: ["CenterOfMass", "GFM_Metals", "NeutralHydrogenAbundance"],
        1: ["SubfindDensity", "SubfindHsml", "SubfindVelDisp"],
        4: ["GFM_Metals"],
    }

    def __init__(self, base_path: str | Path, snapshot: int, halo_index: int, **kwargs):
        """Initialize a TNGGalaxy instance.

        Args:
            base_path (str): Base path to the Illustris(TNG) snapshots.
            snapshot (int): Snapshot ID {0-99}.
            halo_index (int): Halo index, index of the subhalo ID list.
            kwargs (dict): Additional keyword arguments
                - verbose (bool): If True, print information to the command line.
        """
        kwargs.setdefault("verbose", False)
        self.base_path = base_path
        self.snapshot = snapshot
        self._halo_index = halo_index
        self.subhalo_ids = self.subhalo_list(base_path, snapshot, verbose=kwargs["verbose"])
        self.subhalo = self.load_subhalo(self.halo_index, verbose=kwargs["verbose"])
        super().__init__(**kwargs)

    @property
    def halo_index(self):
        """The halo index getter."""
        if hasattr(self, "_halo_index"):
            return self._halo_index
        return 0

    @halo_index.setter
    def halo_index(self, index):
        """The halo index setter."""
        if hasattr(self, "_halo_index") and self._halo_index == index:
            return
        self._halo_index = index

        self.subhalo = self.load_subhalo(index) if hasattr(self, "subhalo") else None
        self.data = (
            self.load_data(as_float32=self.as_float32, verbose=self.verbose)
            if hasattr(self, "data")
            else None
        )

    @SPHGalaxy.p_idx.setter
    def p_idx(self, p_idx: int | str):
        """Setter for particle index."""
        SPHGalaxy.p_idx.fset(self, p_idx)

    def load_header(self) -> dict:
        """Load the header (overrides parent class method).

        Returns:
            (dict): a dictionary containing the headers from a snapshot and the
              corresponding group catalog.
        """
        header = {}
        header["catalog"] = tng.groupcat.load_header(self.base_path, self.snapshot)
        header["snapshot"] = tng.snapshots.load_header(self.base_path, self.snapshot)
        return header

    def load_cosmology(self, cosmo_pars: dict | None = None, in_place: bool = True) -> dict:
        """Load the cosmological parameters from the header dictionary.

        Note: This overrides the parent class method.

        Args:
            cosmo_pars: The default dictionary for the CosmoModel dataclass.
            in_place: Load CosmoModel into instanace property `cosmology` directly.

        Returns:
            (dict): a dictionary with cosmological parameters pulled the header.
        """
        cosmo_pars = {} if cosmo_pars is None else cosmo_pars
        cosmo_pars.setdefault("omega_l", self.header["snapshot"].get("OmegaLambda", 0.6911))
        cosmo_pars.setdefault("omega_m", self.header["snapshot"].get("Omega0", 0.3089))
        cosmo_pars.setdefault("omega_k", self.header["snapshot"].get("OmegaK", 0))
        cosmo_pars.setdefault("z", self.header["snapshot"].get("Redshift", None))
        cosmo_pars.setdefault("z", 1. / self.header["snapshot"].get("Time", 1) - 1)
        cosmo_pars.setdefault("h", self.header["snapshot"].get("HubbleParam", 0.6774))
        if in_place:
            self.cosmology = CosmoModel(**cosmo_pars)
        return cosmo_pars

    @staticmethod
    def subhalo_list(
        base_path: str,
        snapshot: int,
        filtered: bool = True,
        verbose: bool = False,
    ) -> NDArray:
        """Pull a list of subhalo IDs from the snapshot's FOF group catalog.

        Note: the indices of the list are halo IDs.

        Args:
            base_path: Base path to the Illustris(TNG) snapshots.
            snapshot: Snapshot ID {0-99}.
            filtered: If True, negative subhalo IDs are filtered out.
            verbose: If True, print out results to the command line.

        Returns:
            (list | np.ndarray): a list of subhalo i.e. galaxy IDs.
        """
        kwargs = {"fields": ["GroupFirstSub"]}
        if verbose:
            print(f"Searching group catalog: snapshot {snapshot}")
        galaxy_list = tng.groupcat.load_halos(base_path, snapshot, as_array=True, **kwargs)
        galaxy_list = galaxy_list[galaxy_list >= 0] if filtered else galaxy_list
        if verbose:
            N_g = galaxy_list.max()
            N_g += 0 in galaxy_list
            print(f"Found {N_g} groups in catalog: snapshot {snapshot}")
        return galaxy_list

    def load_subhalo(self, halo_index: int = 0, verbose: bool = False) -> dict:
        """Load the FOF subhalo metadata of a specified group ID.

        Args:
            halo_index: the group ID corresponding to a subhalo ID
            verbose: Print status information to the command-line.

        Returns:
            (dict): dictionary containing the FOF subhalo metadata
        """
        if halo_index < 0 or halo_index >= len(self.subhalo_ids):
            halo_index = 0
        sh_id = self.subhalo_ids[halo_index]
        if verbose:
            print(f"Loading group: {sh_id}")
        data = tng.groupcat.load_single(self.base_path, self.snapshot, subhalo_id=sh_id)
        return data

    @property
    def N_particles_type(self) -> list[int]:
        """Number of particles per type in the galaxy."""
        if hasattr(self, "subhalo"):
            return [int(i) for i in self.subhalo["SubhaloLenType"]]
        return []

    @property
    def N_particles(self) -> int:
        """Total number of particles in the galaxy."""
        if hasattr(self, "subhalo"):
            return int(self.subhalo["SubhaloLen"])
        return -1

    @property
    def center(self) -> au.Quantity:
        """Center position getter of the galaxy in units of L/h."""
        if not hasattr(self, "subhalo"):
            self.subhalo = self.load_subhalo(self.halo_index)
        L = self.units("l/h")
        if "SubhaloPos" in self.subhalo:
            return (self.subhalo["SubhaloPos"] * L).to(au.kpc)
        return None

    def load_data(
        self,
        halo_index: int | None = None,
        primary_fields: list[str] | None = None,
        optional_fields: list[str] | None = None,
        particle_type: str | None = None,
        **kwargs,
    ):
        """Load the snapshot's relevant subset data of the specified group ID.

        Args:
            halo_index: Galaxy/Halo ID in the simulation.
            primary_fields: The primary fields to load (less fields load faster)
            optional_fields: The secondary fields to load (less fields load faster)
            particle_type: Manually set the particle type.
            kwargs:
              - verbose (bool): Print information to the command-line.

        Returns:
            (dict): dictionary containing the relevant snapshot data
        """
        kwargs.setdefault("verbose", False)
        verbose = kwargs.pop("verbose")
        if particle_type is not None:
            p_idx = tng.util.pidx_from_ptype(particle_type)
        else:
            p_idx = self.p_idx
            particle_type = self.particle_type
        if primary_fields is None:
            primary_fields = self.primary_hdf5_fields[p_idx]
        if optional_fields is None:
            optional_fields = self.optional_hdf5_fields[p_idx]
        if halo_index is None or halo_index < 0 or halo_index >= len(self.subhalo_ids):
            halo_index = self.halo_index
        subset = tng.snapshots.snapshot_offsets(self.base_path, self.snapshot, halo_index, "Group")
        if verbose:
            print(f"Loading primary fields from snapshot [{particle_type}]: {self.snapshot}")
        data = tng.snapshots.load_snapshot(
            self.base_path,
            self.snapshot,
            particle_type,
            subset=subset,
            fields=primary_fields,
            **kwargs,
        )
        try:
            if verbose:
                print(f"Loading optional fields from snapshot [{particle_type}]: {self.snapshot}")
            data.update(
                tng.snapshots.load_snapshot(
                    self.base_path,
                    self.snapshot,
                    particle_type,
                    subset=subset,
                    fields=optional_fields,
                    as_array=False,
                    **kwargs,
                )
            )
        except Exception as ex:
            if ("Particle type" in ex.args[0]) and ("does not have field" in ex.args[0]):
                for f in optional_fields:
                    data[f] = None
            else:
                raise
        if particle_type == "gas" and ("GFM_Metals" not in data or data["GFM_Metals"] is None):
            data["GFM_Metals"] = np.array([[0.76]])
        if "CenterOfMass" not in data or data["CenterOfMass"] is None:
            data["CenterOfMass"] = data["Coordinates"]
        return data

    @property
    def temperature(self):
        """Temperature data getter in units of Kelvin."""
        # ~ O(1)
        x_H = self.x_H
        # choose constants in favorable units to avoid overflows
        gamma = 5.0 / 3
        m_p = ac.m_p.to(au.u)
        k_B = ac.k_B.to(au.eV / au.K)
        mu = 4 * m_p / (1 + 3 * x_H + 4 * x_H * self.x_e)
        T = (gamma - 1) * mu * self.internal_energy / k_B
        return T.to(au.K)

    @property
    def magnetic_field(self):
        """Magnetic vector field in cgs units of Gauss."""
        if not hasattr(self, "data"):
            self.data = self.load_data()
        srP = self.units("sqrtP")
        if "MagneticField" in self.data:
            return self.data["MagneticField"][:] * srP
        return None

    @property
    def magnetic_field_strength(self):
        """Magnetic field strength in units of SI Gauss."""
        bfield3D = self.magnetic_field * np.sqrt(4 * np.pi * ac.mu0)
        return np.linalg.norm(bfield3D, axis=-1).to(au.Gauss)

    @property
    def radii(self):
        """Particle radius distances assuming spherical shape."""
        # Voronoi cell volume
        vol = self.masses / self.density
        # estimate radii assuming spherical shape
        r_cell = np.power(0.75 * np.pi * vol, 1.0 / 3).to(au.kpc)
        return r_cell

    @property
    def r_cell(self):
        """Alias for radii."""
        return self.radii

    @property
    def hsml(self):
        """Average kernel smoothing length."""
        # hsml has in mind a cubic spline that =0 at h, I think
        # find_fwhm(CubicSplineKernel().kernel)
        hsml = 2.5 * self.r_cell  # * find_fwhm(CubicSplineKernel().kernel)
        return hsml

    def get_mapping_arrays(
        self,
        keys: list[str] | None = None,
        factors: list[float] = None,
        verbose: bool = False,
    ) -> list:
        """Fetch data (arrays or scalars) from a TNGGalaxy object given keys.

        Args:
            keys: Keys to fetch data.
            factors: Factors for modifying the fetched data.
            verbose: If True, print status updates to command line.

        Returns:
            (list): List of fetched data arrays or scalars.
        """
        if keys is None:
            keys = ["particle_positions", "masses", "radii"]
        if factors is None:
            factors = [1, 1, 3]
        vals = []
        for key, f in zip(keys, factors):
            if isinstance(key, tuple | list):
                try:
                    v1 = getattr(self, key[0])
                except Exception:
                    v1 = self.subhalo[key[0]]
                try:
                    v2 = getattr(self, key[1])
                except Exception:
                    v2 = self.subhalo[key[1]]
                v = v1 * v2 * f
            else:
                try:
                    v = getattr(self, key) * f
                except Exception:
                    v = self.subhalo[key] * f
            vals.append(v)
        if verbose:
            print(f"Loading arrays: {keys}...")
        return [*vals]

    def generate_map(
        self,
        keys: list[str] | None = None,
        factors: list[float] | None = None,
        use_half_mass_rad: bool = True,
        fh: float = 3,
        grid_size: int = 512,
        xaxis: int = 0,
        yaxis: int = 1,
        periodic: bool = True,
        assignment_func: Callable = voronoi_RT_2D,
        tracers: int | None = None,
        divisions: int | None = None,
        rot: list[int] | list[float] | None = None,
        verbose: bool = False,
    ) -> tuple[au.Quantity, au.Quantity, int] | Any:
        """Generate raytracing projection map.

        Args:
            keys: Keys to fetch data for projecting onto the map.
            factors: Factors for modifying the projection data.
            use_half_mass_rad:
              If True, the SubhaloHalfmassRad from the subfind catalog is used for
              selecting relevant particles. Otherwise, a fraction of the entire
              particle extent is used.
            fh: Expansion factor for the SPH particle radii.
            grid_size: The size of the maps/images. Default: 512.
            xaxis: Projection axis for x.
            yaxis: Projection axis for y.
            periodic: Use periodic boundary conditions for the projection.
            assignment_func: Mass assignment algorithm; one of
              [voronoi_RT_2D, voronoi_NGP_2D].
            tracers: Number of tracer particles to use for the Nearest Grid Point algorithm.
            divisions: Number of sphere divisions to use for the Nearest Grid Point algorithm.
            rot:
              Angles to rotate the particle positions.
            verbose: If True, print status updates to command line.

        Returns:
            (np.ndarray, np.ndarray, int): The projected map, the map extent, and
               number of particles projected.
        """
        L, M, T = au.Mpc, au.Msun, au.K
        projected = np.zeros((grid_size, grid_size), dtype=np.float64)
        if keys is None:
            keys = ["particle_positions", "masses", "radii", "center"]
        if factors is None:
            factors = [1, 1, fh, 1]
        if use_half_mass_rad:
            if "SubhaloHalfmassRadType" not in keys:
                keys += ["SubhaloHalfmassRadType"]
                factors += [self.units("l/h")]
            else:
                factors.insert(keys.index("SubhaloHalfmassRadType"), self.units("l/h"))
        if assignment_func not in [voronoi_RT_2D, voronoi_NGP_2D]:
            raise ValueError(f"Assignment function `{assignment_func.__name__}` not compatible.")
        uarrs = self.get_mapping_arrays(keys=keys, factors=factors, verbose=verbose)
        if rot is not None:
            rot_op = R.y(rot[1]) * R.x(rot[0])
            uarrs[0] = rot_op(uarrs[0]).astype(np.float32)
            uarrs[3] = rot_op(uarrs[3]).astype(np.float32)
        # Ngids = uarrs[0].shape[0]
        hmr = uarrs[4][0] if use_half_mass_rad else None  # always use gas (p_idx=0) hmr
        # hmr = (uarrs[4][obj.p_idx] if use_half_mass_rad else None)
        idcs, limits = indices_within_box(uarrs[0], uarrs[3], radius=hmr, verbose=verbose)
        hmr2 = limits[3] - limits[0]
        args = strip_ap_units(*uarrs[:3], *limits[:2], hmr2, mask=idcs, units=[L, M, T * M])
        if tracers is not None:
            args.append(tracers)
        else:
            args.append(xaxis)
        if divisions is not None:
            args.append(divisions)
        else:
            args.append(yaxis)
        if verbose:
            print(f"Raytracing particles with `{assignment_func.__name__}`...")
        try:
            assignment_func(projected, *args, periodic, verbose=True)
        except Exception as e:
            print(e)
            return projected * uarrs[1].unit / L**2
        projected = projected * uarrs[1].unit / L**2
        return projected, hmr2 / 2 * np.array([-1, 1, -1, 1]), idcs.shape[0]


class ArepoGalaxy(TNGGalaxy):
    """An Arepo Galaxy parser (alias to TNGGalaxy for now)."""


class GasolineGalaxy(SPHGalaxy):
    """A Gasoline Galaxy parser."""


#     @staticmethod
#     def read(filename, properties: dict = None, **kwargs):
#         """
#         Read and configure a gasoline simulation using pynbody
#
#         Args:
#           filename (str): Relative path to the data of the simulation
#           properties (dict): Extra data properties to be added
#         """
#         if properties is None:
#             properties = {}
#         if os.path.exists(filename):
#             s = filename  # s = pb.load(filename, **kwargs)
#             # for k in properties:
#             #    s.properties[k] = properties[k]
#             # s.physical_units()
#             return s
#         raise FileNotFoundError(
#             f"Problem loading the given file: {filename}\n"
#             f"Current working directory: {os.getcwd()}"
#         )
#
#


def indices_within_box(
    pos: np.ndarray | au.Quantity,
    center: list | np.ndarray | au.Quantity,
    radius: float | au.Quantity = None,
    fraction: float = 1.0,
    verbose: bool = False,
) -> tuple[au.Quantity, list[au.Quantity]]:
    """Get particle indices within a box of given radius from the centre, e.g. half-mass radius.

    Args:
        pos: Particle positions to be filtered.
        center: Center position of the cube.
        radius: Radius, i.e. half-side of the cube.
        fraction: Box fraction for the default radius if not given.
        verbose: If True, print status updates to command line.

    Returns:
        (au.Quantity, list[au.Quantity]):
          Indices of the filtered particle positions and their 3D extent.
    """
    pos_extent = (
        pos[:, 0].max() - pos[:, 0].min(),
        pos[:, 1].max() - pos[:, 1].min(),
        pos[:, 2].max() - pos[:, 2].min(),
    )
    w = max(pos_extent)
    if verbose:
        print(f"Box extent:  {pos_extent[0]}   {pos_extent[1]}   {pos_extent[2]}")
    if radius is None:
        radius = 0.25 * fraction * w
    else:
        radius *= fraction
    if verbose:
        print(f"Radius: {radius}")
    xmin, xmax = center[0] - radius, center[0] + radius
    ymin, ymax = center[1] - radius, center[1] + radius
    zmin, zmax = center[2] - radius, center[2] + radius
    if verbose:
        print(f"Extent: {xmax - xmin}   {ymax - ymin}   {zmax - zmin}")
    indices = np.where(
        (pos[:, 0] > xmin)
        & (pos[:, 0] < xmax)
        & (pos[:, 1] > ymin)
        & (pos[:, 1] < ymax)
        & (pos[:, 2] > zmin)
        & (pos[:, 2] < zmax)
    )[0]
    if verbose:
        print(f"Selected particles: {indices.shape[0]:,} / {pos.shape[0]:,}")
    return indices, [xmin, ymin, zmin, xmax, ymax, zmax]


def strip_ap_units(
    *args,
    mask: list | None = None,
    units: list[au.Unit] | None = None,
    dtype: Any = np.float32,
) -> list:
    """Remove astropy units from data arrays or scalars.

    Args:
        args: Data arrays or scalars with astropy units.
        mask: Mask for the data arrays.
        units: Astropy units to be stripped.
        dtype: Data type of the stripped data array.

    Returns:
        (list[np.ndarray]): Of astropy units stripped data arrays or scalars.
    """
    arg_ls = list(args)
    for i, arg in enumerate(arg_ls):
        if units is not None:
            for u in units:
                if arg_ls[i].unit.is_equivalent(u):
                    arg_ls[i] = arg_ls[i].to(u)
        arg_ls[i] = arg_ls[i].value
        if isinstance(arg_ls[i], np.ndarray):
            if mask is not None:
                arg_ls[i] = arg_ls[i][mask].astype(dtype)
    return arg_ls


if __name__ == "__main__":
    import pprint

    # Illustris loads
    tng_path = "/scratch/data/illustris/tng50-1"
    tng_id = 99
    tng_src = TNGGalaxy(tng_path, tng_id, halo_index=10, as_float32=True)
    pprint.pprint(tng_src.header)
    pprint.pprint(tng_src.cosmology)
    pprint.pprint(tng_src.subhalo.keys())
    pprint.pprint(tng_src.data.keys())
    # dists = tng_src.kd_tree(k=8, threads=4)
    # print(dists.shape, dists)



#!/usr/bin/env python
# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""skais-mapper entry points: configure [or conf|c], generate [or gen|g], help [or h]."""

import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import skais_mapper


def parse_args(return_parser: bool = False, **kwargs):
    """Parse arguments.

    Args:
        return_parser: If True, returns parser instead of parsed argument dictionary.
        kwargs: Additional keyword arguments for compatibility.
    """
    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter, add_help=False)
    parser.add_argument(
        "mode",
        nargs="?",
        choices=[
            "configure",
            "conf",
            "c",
            "generate",
            "gen",
            "g",
            "help",
            "h",
        ],
        help="Available subcommands: use `skais-mapper [subcommand] -h` for details",
    )
    if return_parser:
        return parser
    args, _ = parser.parse_known_args()
    configs = vars(args)
    return configs


def main():
    """Main entry point for skais-mapper."""
    parser = parse_args(return_parser=True)
    args, _ = parser.parse_known_args()
    options = vars(args)
    if options["mode"]:
        sys.argv.pop(1)
        match options["mode"]:
            case "configure" | "conf" | "c":
                skais_mapper.configure.create()
            case "generate" | "gen" | "g":
                skais_mapper.generate.run()
            case _:
                parser.print_help()
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2025-present Philipp Denzel <phdenzel@gmail.com>
# SPDX-FileNotice: Part of skais-mapper
# SPDX-License-Identifier: GPL-3.0-or-later
"""Image/map data readers and (HDF5) writers."""

from pathlib import Path
import warnings
import numpy as np
import h5py
from h5py import Dataset as H5Dataset
from tqdm import tqdm
from PIL import Image
from skais_mapper.utils import nbytes, current_time, next_prime


class ImgRead:
    """Flexible image reader for multiple formats."""

    def __call__(
        self,
        paths: str | Path | list[str | Path] | None = None,
        squash: bool = True,
        pad_val: int | float = 0,
        **kwargs,
    ) -> np.ndarray:
        """Automatically determine file type and read data appropriately.

        Args:
            paths: File path to the image to be read.
            squash: If multiple paths are passed, merge and squash arrays.
            pad_val: Padding value to be used for shape expansion if multiple
              paths are passed and images have different shape (default: 0).
            **kwargs: Additional keyword arguments for parser functions:
              `_read_npy`, `_read_png`, or `_read_jpg`.

        Returns:
            Numpy ndarray of the image data.
        """
        if isinstance(paths, list | tuple):
            data = [self(p, **kwargs) for p in tqdm(paths, desc="ImgRead")]
            if data and squash:
                try:
                    data = np.concatenate(data)
                except ValueError:
                    data = self._stack_max_expand(data, pad_val=0)
            return data
        filepath = Path(paths) if paths is not None else Path("")
        # print("Filepath", filepath)
        match filepath.suffix:
            case ".npy":
                return self._read_npy(filepath, **kwargs)
            case ".jpg":
                return self._read_jpg(filepath, **kwargs)
            case ".png":
                return self._read_png(filepath, **kwargs)
            case _:
                return None

    @staticmethod
    def _stack_max_expand(data: list[np.ndarray], pad_val: int | float = 0):
        """Expand image arrays to maximum shape by padding and stack them.

        Args:
            data: list of arrays to be padded.
            pad_val: Padding value to be used (default: 0).

        Returns:
            The stacked numpy array.
        """
        max_shape = np.max([d.shape for d in data], axis=0)
        padded = []
        for arr in data:
            pad_width = [(0, maxd - currd) for currd, maxd in zip(arr.shape, max_shape)]
            parr = np.pad(arr, pad_width, mode="constant", constant_values=pad_val)
            padded.append(parr)
        return np.concatenate(padded)

    @staticmethod
    def _read_npy(
        path: str | Path,
        dtype: str | np.dtype | None = None,
        expand_dim: bool = True,
        **kwargs,
    ) -> np.ndarray:
        """Numpy file parser.

        Args:
            path: Filepath to the numpy file to be read
            dtype: Typecode or data-type to which the array is cast
            expand_dim: Expand array dimensions with new axis for stacking.
            **kwargs: Additional keyword arguments for `np.load`.

        Returns:
            Numpy ndarray read from numpy file

        Raises:
            (FileNotFoundError): if the input file path does not exist
        """
        filepath = Path(path)
        kwargs.setdefault("mmap_mode", None)
        if not filepath.exists():
            raise FileNotFoundError(f"File {str(filepath)} does not exist.")
        data_array = np.load(filepath, **kwargs)
        if dtype is not None:
            data_array = data_array.astype(dtype)
        if expand_dim and len(data_array.shape) <= 3:
            data_array = data_array[np.newaxis, ...]
        return data_array

    @staticmethod
    def _read_png(
        path: str | Path,
        dtype: str | np.dtype | None = None,
        expand_dim: bool = True,
        **kwargs,
    ) -> np.ndarray:
        """PNG file parser.

        Args:
            path: Filepath to the PNG file to be read.
            dtype: Typecode or data-type to which the array is cast.
            expand_dim: Expand array dimensions with new axis for stacking.
            **kwargs: Additional keyword arguments for `PIL.Image.open`.

        Returns:
            Numpy ndarray read from PNG file

        Raises:
            (FileNotFoundError): if the input file path does not exist
        """
        filepath = Path(path)
        kwargs.setdefault("mode", "r")
        if not filepath.exists():
            raise FileNotFoundError(f"File {str(filepath)} does not exist.")
        with Image.open(filepath, **kwargs) as img:
            data_array = np.asarray(img)
        if dtype is not None:
            data_array = data_array.astype(dtype)
        if expand_dim and len(data_array.shape) <= 3:
            data_array = data_array[np.newaxis, ...]
        return data_array

    @staticmethod
    def _read_jpg(
        path: str | Path,
        dtype: str | np.dtype | None = None,
        expand_dim: bool = True,
        **kwargs,
    ) -> np.ndarray:
        """JPG file parser.

        Args:
            path: Filepath to the JPG file to be read.
            dtype: Typecode or data-type to which the array is cast.
            expand_dim: Expand array dimensions with new axis for stacking.
            **kwargs: Additional keyword arguments for `PIL.Image.open`.

        Returns:
            Numpy ndarray read from JPG file

        Raises:
            (FileNotFoundError): if the input file path does not exist
        """
        return ImgRead._read_png(path, dtype=dtype, expand_dim=expand_dim, **kwargs)


class Img2H5Buffer:
    """Parse images (incrementally or all at once) and write to HDF5 files.

    The directory structure of a dataset should be as follows:
      - <data_dir>: /path/to/dataset/root
      - <file_groups>: image class as a subdirectory in the dataset
      - image file <extensions>: {npy | jpg | png | etc.}
    E.g. file paths of the following structure:
      /path/to/dataset/root/**/image_class/**/423120.npy
    HDF5 files end up being:
      /image_class/dataset

    Note: by default the entire dataset is loaded into cache
    """

    default_target_name = "{}.hdf5"
    extensions = [".jpg", ".png", ".npy"]

    def __init__(
        self,
        path: str | Path = None,
        target: str | Path = None,
        data: np.ndarray | dict = None,
        size: int | float | str = "1G",
    ):
        """Constructor.

        Args:
            path: Path to a data directory where the source files are located.
            target: Filename of the HDF5 file to be written.
            data: Alternative input format to `path` for adding data arrays or
              dictionaries directly to the buffer queue.
            size: Buffer cache size in bytes or passed as string.
        """
        self.files = self.glob_path(path)
        self.queue: list[np.ndarray | dict] = []
        if isinstance(data, np.ndarray | dict):
            self.queue.append(data)
        if target is None:
            target = self.default_target_name.format(current_time()[2:])
        self.target = Path(target)
        self.cache_size = nbytes(size)
        self.index = -1

    @staticmethod
    def _split_glob(
        path: str | list[str], relative: bool = False
    ) -> tuple[list[str], list[str | None]]:
        """Split path containing wildcard into root and wildcard expression."""
        root: list[str]
        file_key: list[str | None]
        if isinstance(path, str) and "*" in path:
            if relative:
                path = path[1:] if path.startswith("/") else path
            components = Path(path).parts
            iwc = [i for i, c in enumerate(components) if "*" in c][0]
            root = [str(Path().joinpath(*components[:iwc]))]
            file_key = [str(Path().joinpath(*components[iwc:]))]
        elif isinstance(path, list | tuple):
            root = []
            file_key = []
            for p in path:
                f, k = Img2H5Buffer._split_glob(p)
                root += f
                file_key += k
        else:
            root = [str(path)]
            file_key = [None]
        return root, file_key

    @staticmethod
    def glob_path(
        path: str | Path | list[str] | list[Path], extensions: str | list[str] = None
    ) -> list[Path]:
        """Glob path recursively for files.

        Args:
            path: Filename, path or list, can contain wildcards `*` or `**`.
            extensions: File extension to look fo
        """
        files: list[Path] = []
        extensions = [extensions] if isinstance(extensions, str) else extensions
        root, file_key = Img2H5Buffer._split_glob(path)
        for p, k in zip(root, file_key):
            if k is None:
                files.append(Path(p))
            else:
                path_files = [
                    f
                    for f in Path(p).rglob(k)
                    if f.is_file() and f.suffix in Img2H5Buffer.extensions
                ]
                files += sorted(path_files)
        files = [p for p in files if p.exists()]
        return files

    @property
    def n_files(self) -> int:
        """Number of files to be parsed."""
        if self.files is not None:
            return len(self.files)
        return 0

    @property
    def nbytes(self) -> list["nbytes"]:
        """List of the number of bytes for each buffer file."""
        if self.files:
            return [nbytes(f.stat().st_size) for f in self.files]
        return [nbytes(0)]

    @property
    def total_nbytes(self) -> list["nbytes"]:
        """Total number of bytes for buffer."""
        return sum(self.nbytes)

    def __str__(self):
        """String representation of the instance."""
        return (
            f"{self.__class__.__name__}("
            f"#b{len(self.queue)}#f{self.n_files}"
            f":{self.total_nbytes.as_str()}"
            f"@{self.cache_size.as_str()}->{self.target.name}"
            ")"
        )

    def configure_rdcc(
        self,
        cache_size: int | float | str | None = None,
        f: int = 10,
        verbose: bool = False,
        **kwargs,
    ) -> dict:
        """Automatically configure HDF5 data chunking for optimal writing.

        Args:
            cache_size: Cache of the entire buffer.
            f: Factor with which to increase the number of slots.
            verbose: Print additional information to stdout.
            **kwargs: Additional keyword arguments such as
              - `rdcc_nbytes`: See h5py.File or below for details.
              - `rdcc_w0`: See h5py.File or below for details.
              - `rdcc_nslots`: See h5py.File or below for details.

        Returns:
            `rdcc_nbytes`: sets the total size (measured in bytes) of the raw data
              chunk cache for each dataset. This should be set to the size of each
              chunk times the number of chunks that are likely to be needed in cache.
            `rdcc_w0`: sets the eviction policy for chunks from the cache when more
              space is needed. 0 is always last used chunk, 1 the last used chunk fully
              read or written, and inbetween values hybrid policies.
            `rdcc_nslots`: is the number of chunk slots in the cache for each dataset.
              In order to allow the chunks to be looked up quickly in cache, each chunk
              is hashed. Thus, it should be large enough to minimize the number of hash
              value collisions. At minimum 10, for maximum performance about 100 times
              larger as the number of chunks which fit in cache, ideally a prime number.
        """
        if cache_size is None:
            cache_size = self.cache_size
        else:
            cache_size = nbytes(cache_size)
        slots_size = max(self.nbytes) if self.nbytes else nbytes("2M")
        n_slots = int(cache_size / slots_size) if self.n_files else 100_000
        # avoid calculating prime numbers if previous configuration looks similar
        if (
            hasattr(self, "_rdcc")
            and self._rdcc["rdcc_nbytes"] == int(cache_size)
            and self._rdcc["rdcc_w0"] == kwargs.get("rdcc_w0", 1.0)
        ):
            return self._rdcc
        kwargs.setdefault("rdcc_nbytes", int(cache_size))
        kwargs.setdefault("rdcc_w0", 1.0)
        kwargs.setdefault("rdcc_nslots", next_prime(int(n_slots * f)))
        if verbose:
            sample_size = max(self.nbytes)
            print(f"Sample size: {sample_size}")
            print(f"Slot size: {nbytes(kwargs['rdcc_nbytes'] / kwargs['rdcc_nslots'])}")
            print(f"Slots: {kwargs['rdcc_nslots']}({n_slots})")
            print(f"Cache size: {nbytes(kwargs['rdcc_nbytes'])}")
            print(f"Eviction policy: {kwargs['rdcc_w0']}")
        self._rdcc = kwargs
        return kwargs

    @staticmethod
    def _h5py_file_kwargs(kwargs, defaults: dict = {}, in_place: bool = True):
        """Extract or filter out relevant keyword arguments for `h5py.File`."""
        file_kw = {}
        for key in defaults:
            file_kw.setdefault(key, defaults[key])
        for key in [
            "driver",
            "userblock_size",
            "track_order",
            "fs_strategy",
            "fs_persist",
            "fs_threshold",
            "fs_page_size",
            "page_buf_size",
            "min_meta_keep",
            "min_raw_keep",
            "locking",
            "alignment_threshold",
            "alignment_interval",
        ]:
            if key in kwargs:
                if in_place:
                    file_kw[key] = kwargs.pop(key)
                else:
                    file_kw[key] = kwargs.get(key)
        if in_place:
            return file_kw, kwargs
        return file_kw

    @staticmethod
    def _h5py_create_dataset_kwargs(kwargs, defaults: dict = {}, in_place: bool = True):
        """Extract or filter out relevant keyword arguments for `h5py.Group.create_dataset`."""
        create_dataset_kw = {}
        for key in defaults:
            create_dataset_kw.setdefault(key, defaults[key])
        for key in [
            "shape",
            "dtype",
            "chunks",
            "maxshape",
            "compressioncompression_opts",
            "scaleoffset",
            "shuffle",
            "fletcher32",
            "fillvalue",
            "fill_time",
            "track_times",
            "track_order",
            "external",
            "allow_unknown_filter",
        ]:
            if key in kwargs:
                if in_place:
                    create_dataset_kw[key] = kwargs.pop(key)
                else:
                    create_dataset_kw[key] = kwargs.get(key)
        if in_place:
            return create_dataset_kw, kwargs
        return create_dataset_kw

    @property
    def page(self) -> np.ndarray | dict | None:
        """Buffer page ready to be written to file."""
        if self.queue:
            return self.queue[0]
        elif self.files:
            return ImgRead()(self.files[0])
        return None

    def store(
        self, data: np.ndarray | dict | str | Path | list[str | Path], squash: bool = True
    ) -> "Img2H5Buffer":
        """Insert data into the buffer queue.

        Args:
            data: Data to be stored in buffer.
            squash: Squash data dimensions if buffer data is compatible.
        """
        if (
            squash
            and isinstance(data, np.ndarray)
            and self.queue
            and isinstance(self.queue[-1], np.ndarray)
        ):
            try:
                self.queue[-1] = np.concatenate((self.queue[-1], data))
            except ValueError:
                self.queue[-1] = ImgRead._stack_max_expand([self.queue[-1], data])
        elif isinstance(data, np.ndarray | dict):
            self.queue.append(data)
        elif isinstance(data, str | Path) or (
            isinstance(data, list) and isinstance(data[0], str | Path)
        ):
            return self.store(ImgRead()(data))
        return self

    def send(self, clear: bool = True) -> np.ndarray | dict | None:
        """Grab first data page from the buffer queue."""
        if self.queue:
            if clear:
                return self.queue.pop(0)
            return self.page
        elif self.files:
            if clear:
                self.store(self.files.pop(0))
            else:
                self.store(self.files[0])
            return self.send(clear=clear)
        return None

    def flush(self) -> np.ndarray | dict | list[np.ndarray | dict] | None:
        """Send all data pages from the buffer queue."""
        if self.files:
            self.store(self.files)
            self.files = []
        if self.queue:
            data = self.queue
            self.queue = []
            if len(data) == 1:
                return data[0]
            return data
        return None

    def inc_write(
        self,
        path: str | Path | None = None,
        group: str = "images",
        data: np.ndarray | dict | None = None,
        expand_dim: bool = True,
        axis: int = 0,
        overwrite: bool | int | None = None,
        verbose: bool = False,
        **kwargs,
    ):
        """Incrementally (append mode) write the buffer to HDF5 file.

        Args:
            path: Filename of the HDF5 file and optionally the path of the HDF5
              group where the dataset is saved.
            group: HDF5 group where to save the dataset. If it does not exist,
              it is created.
            data: Data to be written to the hdf5 file. If None, all files in
              the buffer are written to HDF5 file.
            expand_dim: Expand dimension of data array for stacking.
            axis: Axis of the n-dimensional array where to append
            overwrite: If data should overwrite indices in a pre-existing HDF5 dataset,
              set to the index.
            verbose: Print additional information to stdout.
            kwargs: Additional keyword arguments for `Img2H5Buffer.configure_rdcc`,
              h5py.File`, and/or `h5py.Group.create_dataset`.
        """
        if path is None:
            path = self.target
        else:
            path = Path(path)
        if ":" in str(path):
            group = str(path).split(":")[1]
            path = Path(str(path).split(":")[0])
        if group.startswith("/"):
            group = group[1:]
        if group is None:
            raise ValueError("inc_write: no data target, add valid `group`...")
        if data is None:
            data = self.flush()
        elif isinstance(data, np.ndarray) and expand_dim:
            data = data[np.newaxis, ...]
        if isinstance(overwrite, bool):
            overwrite = 0 if overwrite else None
        # configure HDF5 chunk caching
        rdcc = {"cache_size": self.cache_size, "verbose": verbose}
        for key in ["f", "rdcc_nbytes", "rdcc_w0", "rdcc_nslots"]:
            if key in kwargs:
                rdcc[key] = kwargs.pop(key)
        rdcc = self.configure_rdcc(**rdcc)
        # write metadata
        if isinstance(data, dict):
            file_kwargs, kwargs = self._h5py_file_kwargs(
                kwargs,
                defaults={"mode": "a", "libver": "latest"},
            )
            with h5py.File(path, **file_kwargs, **rdcc) as h5:
                g = h5.create_group(f"/{group}") if group not in h5.keys() else h5[group]
                for key in data:
                    g.attrs[key] = data[key]
                if verbose:
                    print(
                        f"Data attribute(s) {tuple(data.keys())} have been "
                        f"written to HDF5 file@[/{group}]"
                    )
        # write data
        elif isinstance(data, np.ndarray):
            file_kwargs, kwargs = self._h5py_file_kwargs(
                kwargs,
                defaults={
                    "mode": "a",
                    "libver": "latest",
                },
            )
            kwargs, _ = self._h5py_create_dataset_kwargs(
                kwargs | file_kwargs,
                defaults={
                    "compression": "gzip",
                    "shuffle": True,
                    "track_times": True,
                    "dtype": data.dtype,
                    "shape": data.shape,
                    "maxshape": data.shape[:axis] + (None,) + data.shape[axis + 1 :],
                    "chunks": data.shape[:axis] + (1,) + data.shape[axis + 1 :],
                },
            )
            with h5py.File(path, **file_kwargs, **rdcc) as h5:
                ds_existed = isinstance(h5[group], H5Dataset) if group in h5 else False
                ds = h5.require_dataset(group, **kwargs, **rdcc)
                ds_samples = ds.shape[axis]
                data_samples = data.shape[axis]
                if not ds_existed:
                    self.index = 0
                elif overwrite is None:
                    self.index = ds_samples
                    ds.resize(self.index + data_samples, axis=axis)
                else:
                    self.index = overwrite
                    if data_samples > ds_samples:
                        ds.resize(self.index + data_samples, axis=axis)
                slc = [slice(None)] * len(ds.shape)
                slc[axis] = slice(self.index, self.index + data_samples)
                ds[tuple(slc)] = data
                if verbose:
                    print(
                        f"Data {data.shape} have been written to HDF5 dataset "
                        f"{ds.shape}@({self.index}:{self.index + data_samples})"
                    )
        else:
            warnings.warn(
                "Img2H5Buffer did not write data to file (either "
                "because the buffer was empty or data was incompatible)."
            )

    def write(
        self,
        path: str | Path | None = None,
        group: str = "images",
        data: np.ndarray | dict | None = None,
        verbose: bool = False,
        **kwargs,
    ):
        """Write all files in buffer to a new HDF5 file.

        Args:
            path: Filename of the HDF5 file and optionally the path of the HDF5
              group where the dataset is saved separated by a colon,
              e.g. `'/path/to/file.hdf5:/path/to/group'`.
            group: HDF5 group where to save the dataset. If it does not exist,
              it is created.
            data: Data to be written to the HDF5 file.
            verbose: Print additional information to stdout.
            kwargs: Additional keyword arguments for `h5py.Group.create_dataset`.
        """
        self.inc_write(path, group=group, data=data, verbose=verbose, **kwargs)


if __name__ == "__main__":
    # test Img2H5Buffer
    for img2h5 in [
        Img2H5Buffer(
            target="/scratch/data/illustris/tng50-1.2D/test_250601_empty.hdf5",
            data=np.random.rand(100, 512, 512),
        ),
        Img2H5Buffer(
            "/scratch/data/illustris/tng50-1.2D/099/dm/dm_tng50-1.99.gid.0003*.png",
            "/scratch/data/illustris/tng50-1.2D/test_250601_png.hdf5",
        ),
        Img2H5Buffer(
            "/scratch/data/illustris/tng50-1.2D/099/dm/**/dm_tng50-1.99.gid.0003*.npy",
            "/scratch/data/illustris/tng50-1.2D/test_250601_npy.hdf5",
            data=np.random.rand(1, 512, 512),
        ),
    ]:
        print()
        print(img2h5)
        g = f"{img2h5.target}:/dm/images"
        img2h5.inc_write(g, f=100, verbose=True, track_order=True)
        # pprint.pprint(img2h5.files[:12])
    print()
    img2h5.inc_write(
        "/scratch/data/illustris/tng50-1.2D/test_250601_npy.hdf5:/dm/images",
        data=np.ones((1, 512, 512)),
        verbose=True,
    )
    img2h5.inc_write(
        "/scratch/data/illustris/tng50-1.2D/test_250601_npy.hdf5:/dm/attrs",
        data={"metadata": 100, "method": "inc_write"},
        verbose=True,
    )

    # clean up
    if 0:
        for f in [
            "/scratch/data/illustris/tng50-1.2D/test_250601_empty.hdf5",
            "/scratch/data/illustris/tng50-1.2D/test_250601_png.hdf5",
            "/scratch/data/illustris/tng50-1.2D/test_250601_npy.hdf5",
        ]:
            Path(f).unlink(missing_ok=True)



"""Setup file for compiling and building Cython/C-extensions."""
import sys
import argparse
from setuptools import Extension, setup
import numpy

parser = argparse.ArgumentParser(
    prog="python setup.py",
    formatter_class=argparse.RawTextHelpFormatter,
    description="Build skais-mapper extensions.\n"
    "Use `python setup.py build_ext --inplace`\n"
    "for building Cython extensions from their corresponding C files or\n"
    "`python setup.py build_ext --inplace --use_cython`\nfor building extensions"
    " from the cython files directly.",
)
parser.add_argument(
    "--use_cython",
    "--cython",
    action="store_true",
    help="Build extensions from their cython files instead of the C files.",
)
parser.add_argument(
    "-a", "--report", action="store_true", help="Generate annotated HTML compilation report."
)
parser.add_argument("build_c", nargs="?", default="", help="Compile cython files and exit.")

args, unk = parser.parse_known_args()
USE_CYTHON = args.use_cython
BUILD_C = args.build_c == "build_c"

ext = ".pyx" if USE_CYTHON or BUILD_C else ".c"
extensions = [
    Extension(
        name="skais_mapper.raytrace",
        sources=["skais_mapper/raytrace" + ext],
        include_dirs=[numpy.get_include()],
        define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
    )
]

if USE_CYTHON or BUILD_C:
    from Cython.Build import cythonize

    extensions = cythonize(extensions, language_level=3, annotate=True)
if BUILD_C:
    exit(0)

for c in ["--cython", "--use_cython", "-a", "--report", "build_c"]:
    if c in sys.argv:
        sys.argv.remove(c)

setup(ext_modules=extensions)
